import unittest
from unittest.mock import patch, MagicMock, mock_open
import urllib.error
from urllib.error import URLError, HTTPError
import socket

from mth5.clients.make_mth5 import FDSN
from pyspedas.mth5.load_fdsn import load_fdsn
from pyspedas.mth5.utilities import _list_of_fdsn_channels
from pyspedas.mth5.utilities import datasets

# This one is unused
from pyspedas.mth5.config import CONFIG

import pyspedas
import pyspedas

import time
import os
import io

import h5py
import pandas as pd

from pyspedas.mth5.config import CONFIG

import loguru
from contextlib import contextmanager

# Handling exceptions of bad data input
from obspy.clients.fdsn.header import FDSNBadRequestException
from obspy.clients.fdsn.header import FDSNNoServiceException


@contextmanager
def loguru_capture_logs(level="INFO", format="{level}:{name}:{message}"):
    """Capture loguru-based logs. There is no other way to assert the loguru log. The custom filter also must be defined here"""
    loguru.logger.remove()
    output = []
    handler_id = loguru.logger.add(output.append, level=level, format=format,
                                   filter=pyspedas.mth5._disable_loguru_warnings)
    yield output
    loguru.logger.remove(handler_id)


class Timer:
    """
    Class that tracks the function call time
    """

    def __enter__(self):
        self.start = time.perf_counter()
        return self

    def __exit__(self, *args):
        self.end = time.perf_counter()
        self.interval = self.end - self.start


class TestMTH5LoadFDSN(unittest.TestCase):
    # Flag to check if there is any open instances of h5 files
    H5OPEN = False

    # This flag result in false if basic test fails and will not run other tests
    BASIC_TEST_FAILED = False

    @classmethod
    def setUpClass(cls):
        # TODO: H5OPEN - is not evaluated correctly. The  @unittest.skipIf decorator evaluates the condition at import time
        # Test if there is any open instances of h5 files
        # This case may not work because tests are executed in a different console where h5 reference may not exist.
        o = h5py.h5f.get_obj_ids(types=h5py.h5f.OBJ_FILE)
        cls.H5OPEN = len(o) > 0

    def setUp(self):
        if len(pyspedas.tplot_names()) > 0:
            pyspedas.del_data('*')

    def test01_load_fdsn_notrange(self):
        """
        Testing the load_fdsn function without a trange.
        The log message should be printed.
        """
        with self.assertLogs(logger=pyspedas.logger, level='ERROR') as cm:
            load_fdsn(network="4P", station="REU49")
        self.assertIn('trange not specified', cm.output[0])
        self.assertFalse('fdsn_4P_REU49' in pyspedas.tnames())

    def test01_load_fdsn_nonetwork(self):
        """
        Test loading data without specifying a network
        The log message should be printed.
        """
        date_start = '2015-06-22T01:45:00'
        date_end = '2015-06-22T02:20:00'

        with self.assertLogs(logger=pyspedas.logger, level='ERROR') as cm:
            load_fdsn(station='REU49', trange=[date_start, date_end])
        self.assertIn('Network not specified', cm.output[0])
        self.assertFalse('fdsn_4P_REU49' in pyspedas.tnames())

    def test01_load_fdsn_nostation(self):
        """
        Test loading data without specifying a station
        The log message should be printed.
        """
        date_start = '2015-06-22T01:45:00'
        date_end = '2015-06-22T02:20:00'

        with self.assertLogs(logger=pyspedas.logger, level='ERROR') as cm:
            load_fdsn(network='4P', trange=[date_start, date_end])
        self.assertIn('Station not specified', cm.output[0])
        self.assertFalse('fdsn_4P_REU49' in pyspedas.tnames())

    def test02_load_fdsn_basic(self):
        date_start = '2015-06-22T01:45:00'
        date_end = '2015-06-22T02:20:00'
        load_fdsn(network="4P", station="REU49", trange=[date_start, date_end])
        load_fdsn(network="4P", station="GAW50", trange=[date_start, date_end])

        try:
            self.assertTrue('fdsn_4P_REU49' in pyspedas.tnames())
            self.assertTrue('fdsn_4P_GAW50' in pyspedas.tnames())
            self.BASIC_TEST_FAILED = False
        except AssertionError:
            self.BASIC_TEST_FAILED = True

    # TODO: this should not work, because BASIC_TEST_FAILED should be evaluated before the call of the test...
    @unittest.skipIf(BASIC_TEST_FAILED, "Basic test failed.")
    # @patch('sys.stdout', new_callable=StringIO)
    def test03_load_fdsn_nowarnings(self):
        """
        Test with and without nowarning flag
        """
        date_start = '2015-06-22T01:45:00'
        date_end = '2015-06-22T02:20:00'

        # Try loading data, which should result in warnings by MTH5
        with loguru_capture_logs(level="WARNING") as output:
            load_fdsn(network="4P", station="REU49", trange=[date_start, date_end], nowarnings=False)
            self.assertTrue(len(output) > 0)
        iswaring = any('WARNING:' in s for s in output)

        if not iswaring:
            self.skipTest("load_fdsn does not produce warnings.")
        with loguru_capture_logs() as output:
            load_fdsn(network="4P", station="REU49", trange=[date_start, date_end], nowarnings=True)
        self.assertFalse(any('WARNING:' in s for s in output))

    @unittest.skipIf(BASIC_TEST_FAILED, "Basic test failed.")
    def test03_load_fdsn_nodownload(self):
        """
        If nodownload flag is specify the h5 file will be reused.
        Testing if the function works faster. Skipping the test if the file was not created.
        """

        date_start = '2015-06-22T01:45:00'
        date_end = '2015-06-22T02:20:00'
        with Timer() as t1:
            load_fdsn(network="4P", station="REU49", trange=[date_start, date_end], nowarnings=False, nodownload=False)

        # Check if the h5 file exist
        if not os.path.isfile(os.path.join(CONFIG['local_data_dir'], '4P_REU49_20150622T014500_20150622T022000.h5')):
            self.skipTest("load_fdsn does not produce h5 file.")

        with Timer() as t2:
            load_fdsn(network="4P", station="REU49", trange=[date_start, date_end], nowarnings=False, nodownload=True)

        self.assertTrue(t1.interval > t2.interval)

    def test04_load_fdsn_samedates(self):
        """
              If dates are the same, tplot variable should not be created.
              load_fdsn should return none
        """
        # The server response has changed to '422: Unprocessable Entity' for this request
        # because of the same start/end times.  So we need to catch the exception now.
        from obspy.clients.fdsn.header import FDSNException

        date_start = '2015-06-22T01:45:00'
        date_end = date_start
        tvar = None

        try:
            tvar = load_fdsn(network="4P", station="REU49", trange=[date_start, date_end])
        except FDSNException as e:
            pass

        self.assertIsNone(tvar)

    @unittest.skipIf(BASIC_TEST_FAILED, "Basic test failed.")
    def test05_load_fdsn_timeclip(self):
        """
        Test if the time is clipped correctly.
        """
        date_start = '2015-06-22T01:45:00'
        date_end = '2015-06-22T02:20:00'

        tvar = load_fdsn(network="4P", station="REU49", trange=[date_start, date_end])
        time, data = pyspedas.get_data(tvar)

        t1 = pyspedas.time_datetime(time[0]).strftime('%Y-%m-%dT%H:%M:%S')
        t2 = pyspedas.time_datetime(time[-1]).strftime('%Y-%m-%dT%H:%M:%S')
        import mth5
        print('mth5 version: ',mth5.__version__)
        import mt_metadata
        print('mt-metadata version: ', mt_metadata.__version__)
        import obspy
        print('obspy version: ', obspy.__version__)

        print('date_start: ', date_start)
        print('t1: ', t1)
        print('date_end: ', date_end)
        print('t2: ', t2)
        self.assertTrue(t1 == date_start)
        self.assertTrue(t2 == date_end)

    def test06_load_fdsn_wrong_network_station(self):
        """
            Test incorrect input. load_fdsn should return raise an error and the temporary files should not exist
        """

        def test_network_station_error(network, station):
            date_start = '2015-06-22T01:45:00'
            date_end = '2015-06-22T02:20:00'

            with self.assertLogs(logger=pyspedas.logger, level='ERROR') as log:
                with self.assertRaises(Exception):
                    load_fdsn(network=network, station=station, trange=[date_start, date_end])
                self.assertIn("HTTP Status code: 400", log.output[0])

            tmp_file = os.path.join(CONFIG['local_data_dir'], f'{network}_{station}.h5')
            self.assertFalse(os.path.isfile(tmp_file))

            with self.assertLogs(logger=pyspedas.logger, level='ERROR') as cm:
                try:
                    load_fdsn(network=network, station=station, trange=[date_start, date_end])
                except:
                    pass
            self.assertIn('Cannot initialize mth5 object', cm.output[0])
            self.assertFalse('fdsn_4P_REU49' in pyspedas.tnames())

        # incorrect network
        test_network_station_error(network="4P_", station="REU49")

        # incorrect station
        test_network_station_error(network="4P", station="REU49_")

    # Mock FDSN and MTH so we do not make unnessesary calls
    @patch('mth5.clients.make_mth5.FDSN.__init__', return_value=None)
    @patch('mth5.mth5.MTH5.__init__', return_value=None)
    def test07_print_request(self, mock_FDSN, mock_MTH5):
        """Test the print_request parameter and ensure print output is correct."""
        date_start = '2015-06-22T01:45:00'
        date_end = '2015-06-22T02:20:00'

        # Terminate function after print is complete
        def side_effect(*args, **kwargs):
            raise SystemExit("Exiting test.")

        # Assign the side effect to the mock objects
        mock_FDSN.side_effect = side_effect
        mock_MTH5.side_effect = side_effect

        # Capture the printed output
        printed_output = io.StringIO()
        with self.assertRaises(SystemExit) as cm:
            with patch('sys.stdout', new=printed_output):
                load_fdsn(trange=[date_start, date_end], network="4P", station="REU49", print_request=True)

        # Check the content of the print output
        printed_df = printed_output.getvalue()
        expected_columns = ['network', 'station', 'start', 'end']
        self.assertTrue(all(col in printed_df for col in expected_columns))

    @patch('mth5.clients.make_mth5.FDSN.make_mth5_from_fdsn_client')
    def test08_request_df_parameter(self, mock_make_mth5):
        """Test the request_df parameter with a custom DataFrame."""
        date_start = '2015-06-22T01:45:00'
        date_end = '2015-06-22T02:20:00'

        custom_request_df = pd.DataFrame({
            "network": ["4P"],
            "station": ["REU49"],
            "location": ["--"],
            "channel": ["*F*"],
            "start": [date_start],
            "end": [date_end]
        })

        def side_effect(*args, **kwargs):
            raise SystemExit("make_mth5_from_fdsn_client called, exiting test.")

        mock_make_mth5.side_effect = side_effect

        with self.assertRaises(SystemExit) as cm:
            load_fdsn(trange=[date_start, date_end], network="4P", station="REU49", request_df=custom_request_df)

        # Verify that the reason for exit is the make_mth5_from_fdsn_client call
        self.assertEqual(str(cm.exception), "make_mth5_from_fdsn_client called, exiting test.")

        # Verify that make_mth5_from_fdsn_client was called with the custom DataFrame
        # mock_make_mth5.assert_called_once_with(custom_request_df, interact=False, path=CONFIG['local_data_dir'])
        self.assertTrue(any(call_args[0][0].equals(custom_request_df) for call_args in mock_make_mth5.call_args_list))

    def test09_invalid_date_or_data(self):
        """Test the function with invalid date formats in trange."""
        invalid_date_start = '2015-06-22'
        invalid_date_end = '21-06-2015'  # Invalid format (should handle ok) but the date end before start should results in Exception
        date_start = '2015-06-22T01:45:00'
        date_end = '2015-06-22T02:20:00'

        with self.assertRaises(FDSNBadRequestException) as cm:
            load_fdsn(trange=[invalid_date_start, invalid_date_end], network="4P", station="REU49")

        # Verify the exception type
        self.assertTrue(isinstance(cm.exception, FDSNBadRequestException))

        with self.assertRaises(FDSNBadRequestException) as cm:
            load_fdsn(trange=[date_start, date_end], network="4P", station="NON_EXISTENT")

        # Verify the exception type
        self.assertTrue(isinstance(cm.exception, FDSNBadRequestException))

        with self.assertRaises(FDSNBadRequestException) as cm:
            load_fdsn(trange=[date_start, date_end], network="NON_EXISTENT", station="REU49")

        # Verify the exception type
        self.assertTrue(isinstance(cm.exception, FDSNBadRequestException))

    # TODO: FOR SOME REASON, this test must be called before test02_load_fdsn_basic. Otherwise it does not work.
    @patch('urllib.request.OpenerDirector.open')
    def test02__connection_errors(self, mock_open):
        """Test the function handling of network connection issues."""
        date_start = '2015-06-22T01:45:00'
        date_end = '2015-06-22T02:20:00'

        # Simulate a no internet connection
        mock_open.side_effect = URLError("Network unreachable")

        with self.assertRaises(Exception) as cm:
            load_fdsn(trange=[date_start, date_end], network="4P", station="REU49")
            print(cm.exception)

        # Verify the exception type and message for URLError
        self.assertTrue(isinstance(cm.exception, FDSNNoServiceException) or isinstance(cm.exception, URLError))
        # self.assertTrue(isinstance(cm.exception, URLError))

    # This test seems to be obsolete
    @unittest.skipIf(H5OPEN, "Open h5 files detected. Close all the h5 references before runing this test")
    @unittest.skip
    def test99_load_fdsn_h5_file_error(self):
        # Testing if the file can be created when error occurs

        request_df = pd.DataFrame(
            {
                "network": ["4P"],
                "station": ["ALW48"],
                "location": ["--"],
                "channel": ["*F*"],
                "start": ["2015-06-22T00:00:00"],
                "end": ["2015-06-24T00:00:00:00"]  # intentional error
            }
        )

        fdsn_object = FDSN(mth5_version='0.2.0', client="IRIS")
        try:
            # This will raise an error because the date format is wrong. But the h5 files will be created and open
            fdsn_object.make_mth5_from_fdsn_client(request_df, interact=False, path=CONFIG['local_data_dir'])
        except Exception as e:
            # Check if the type error is raised
            self.assertIsInstance(e, TypeError)
            pass

        # test is load_fdsn closes the file when error is raised and message is logged
        with self.assertLogs(logger=pyspedas.logger, level="INFO") as cm:
            try:
                load_fdsn(network="4P", station="ALW48", trange=['2015-06-22', '2015-06-24'])
            except Exception as e:
                pass
            # test if we have ERROR and INFO in the log messages
            self.assertTrue(any(map(lambda x: x.find("ERROR") == 0, cm.output)))
            self.assertTrue(any(map(lambda x: x.find("INFO") == 0, cm.output)))

        # Empty lists/dicts evaluate to False
        self.assertFalse(h5py.h5f.get_obj_ids(types=h5py.h5f.OBJ_FILE))

        # import h5py
        # print(h5py.h5f.get_obj_ids(types=h5py.h5f.OBJ_FILE))
        #
        # with self.assertLogs(level="ERROR") as cm:
        #     try:
        #         load_fdsn(network="4P", station="ALW48", trange=['2015-06-22', '2015-06-24'])
        #     except:
        #         pass
        # print(cm)

        pass

    @unittest.skip
    def test99_load_fdsn_no_data(self):
        pass

    def tearDown(self):
        pass


class TestListOfChannels(unittest.TestCase):
    def test01_list_of_channels_type(self):
        """
        Test that the _list_of_channels function returns a string.
        """
        result = _list_of_fdsn_channels()
        self.assertIsInstance(result, str)

    def test02_list_of_channels_divisible_by_three(self):
        """
        Test that the number of channel names (elements) in the list is divisible by 3.
        """
        result = _list_of_fdsn_channels()
        channel_list = result.split(',')
        self.assertEqual(len(channel_list) % 3, 0)

    def test03_list_of_channels_contains_F(self):
        """
        Test that all channel names in the list have "F" in the middle.
        """
        result = _list_of_fdsn_channels()
        channel_list = result.split(',')
        for channel in channel_list:
            # Assuming the format is always one character for band, 'F' for instrument, and one character for orientation
            self.assertTrue(channel[1] == 'F')


class TestDatasetsFunction(unittest.TestCase):
    # Mock data of the responce
    mock_data = """Network|Station|Location|Channel|Latitude|Longitude|Elevation|Depth|Azimuth|Dip|SensorDescription|Scale|ScaleFreq|ScaleUnits|SampleRate|StartTime|EndTime
4P|ALW48||LFE|0|0|0|0|0|0|NIMS|1|1|nT|1|2015-06-18T15:00:36.0000|2015-07-09T13:45:10.0000
4P|ALW48||LFN|0|0|0|0|0|0|NIMS|1|1|nT|1|2015-06-18T15:00:36.0000|2015-07-09T13:45:10.0000
4P|ALW48||LFZ|0|0|0|0|0|0|NIMS|1|1|nT|1|2015-06-18T15:00:36.0000|2015-07-09T13:45:10.0000
"""

    # Set to True if we can get to https://service.iris.edu
    PING_IRIS = False

    @classmethod
    def setUpClass(cls):
        host = "service.iris.edu"
        try:
            # Create a socket object
            with socket.create_connection((host, 443), timeout=5) as sock:
                # If the connection was successful, close it and return accessible
                cls.PING_IRIS = True
        except socket.error as e:
            print(e)
            pass

    def test01_datasets_no_trange(self):
        """Test datasets function with no trange."""
        result = datasets()
        self.assertIsNone(result)

    @patch('urllib.request.urlopen')
    @patch('pyspedas.logger.error')  # Assuming pyspedas.logger.error is the correct path
    def test02_datasets_network_failure(self, mock_logger_error, mock_urlopen):
        """Test datasets function with a network failure."""
        # Simulate an HTTPError, which is a subclass of URLError and has a 'code' attribute.
        url = 'http://example.com'
        error_message = 'Test Error'
        mock_urlopen.side_effect = urllib.error.HTTPError(url, 500, error_message, None, None)

        result = datasets(["2019-11-14", "2019-11-15"])
        mock_logger_error.assert_called_once()
        self.assertTrue(any("HTTP or URL Error" in call_args[0][0] for call_args in mock_logger_error.call_args_list))

        self.assertEqual(result, {})

    @patch('urllib.request.urlopen')
    def test02_datasets_no_data(self, mock_urlopen):
        """Test datasets function with no data returned from network."""
        mock_response = MagicMock()
        mock_response.status = 404
        mock_urlopen.return_value.__enter__.return_value = mock_response
        result = datasets(["2019-11-14", "2019-11-15"])
        self.assertEqual(result, {})

    @patch('urllib.request.urlopen')
    def test04_datasets_valid_data(self, mock_urlopen):
        """Test datasets function with valid data."""

        mock_response = MagicMock()
        mock_response.status = 200
        mock_response.read.return_value = self.mock_data.encode('utf-8')
        mock_urlopen.return_value.__enter__.return_value = mock_response

        expected = {'4P': {'ALW48': {('2015-06-18T15:00:36.0000', '2015-07-09T13:45:10.0000'): ['LFE', 'LFN', 'LFZ']}}}

        result = datasets(["2015-06-22", "2015-06-23"], network="4P", station="ALW48")
        self.assertEqual(result, expected)

    @patch('urllib.request.urlopen')
    def test05_datasets_incorrect_data(self, mock_urlopen):
        """Test datasets function with data which does not have all 3 channels."""
        mock_data = self.mock_data.splitlines()
        mock_data[-2] = "4P|ALW48||UFZ|0|0|0|0|0|0|NIMS|1|1|nT|1|2015-06-18T15:00:36.0000|2015-07-09T13:45:10.0000"
        mock_data = "\n".join(mock_data)

        #         """Network|Station|Location|Channel|Latitude|Longitude|Elevation|Depth|Azimuth|Dip|SensorDescription|Scale|ScaleFreq|ScaleUnits|SampleRate|StartTime|EndTime
        # 4P|ALW48||LFE|0|0|0|0|0|0|NIMS|1|1|nT|1|2015-06-18T15:00:36.0000|2015-07-09T13:45:10.0000
        # 4P|ALW48||LFN|0|0|0|0|0|0|NIMS|1|1|nT|1|2015-06-18T15:00:36.0000|2015-07-09T13:45:10.0000
        # 4P|ALW48||UFZ|0|0|0|0|0|0|NIMS|1|1|nT|1|2015-06-18T15:00:36.0000|2015-07-09T13:45:10.0000
        # """
        mock_response = MagicMock()
        mock_response.status = 200
        mock_response.read.return_value = mock_data.encode('utf-8')
        mock_urlopen.return_value.__enter__.return_value = mock_response

        expected = {}

        result = datasets(["2015-06-22", "2015-06-23"], network="4P", station="ALW48")
        self.assertEqual(result, expected)

    @patch('urllib.request.urlopen')
    def test06_url_construction_with_network(self, mock_urlopen):
        """Test the URL construction when a specific network is provided."""
        datasets(["2015-06-22", "2015-06-23"], network="4P")
        # Extract the URL passed to urlopen
        called_url = mock_urlopen.call_args[0][0]
        self.assertIn("net=4P", called_url)

    @patch('urllib.request.urlopen')
    def test06_url_construction_with_station(self, mock_urlopen):
        """Test the URL construction when a specific station is provided."""
        datasets(["2015-06-22", "2015-06-23"], station="ALW48")
        # Extract the URL passed to urlopen
        called_url = mock_urlopen.call_args[0][0]
        self.assertIn("sta=ALW48", called_url)

    @patch('urllib.request.urlopen')
    def test06_url_construction_with_network_and_station(self, mock_urlopen):
        """Test the URL construction when both network and station are provided."""
        datasets(["2015-06-22", "2015-06-23"], network="4P", station="ALW48")
        # Extract the URL passed to urlopen
        called_url = mock_urlopen.call_args[0][0]
        self.assertIn("net=4P", called_url)
        self.assertIn("sta=ALW48", called_url)

    def test07_datasets_realcall(self):
        """Test datasets function with actual internet connection."""
        # Dynamically decide to skip the test based on the runtime condition
        if not self.PING_IRIS:
            self.skipTest("Cannot reach iris website")

        expected = {'4P': {'ALW48': {('2015-06-18T15:00:36.0000', '2015-07-09T13:45:10.0000'): ['LFE', 'LFN', 'LFZ']}}}
        result = datasets(["2015-06-22", "2015-06-23"], network="4P", station="ALW48")
        self.assertEqual(result, expected)

    @patch('urllib.request.urlopen')
    def test08_datasets_USAarea(self, mock_urlopen):
        """Test datasets function with USAarea set to True or False."""
        mock_response = MagicMock()
        mock_response.status = 200
        mock_response.read.return_value = self.mock_data.encode('utf-8')
        mock_urlopen.return_value.__enter__.return_value = mock_response

        result = datasets(["2015-06-22", "2015-06-23"], network="4P", station="ALW48", USAarea=True)

        # Extract the URL passed to urlopen
        called_url = mock_urlopen.call_args[0][0]
        self.assertIn("maxlat=49", called_url)
        self.assertIn("minlon=-127", called_url)
        self.assertIn("maxlon=-59", called_url)
        self.assertIn("minlat=24", called_url)

        result = datasets(["2015-06-22", "2015-06-23"], network="4P", station="ALW48")

        # Extract the URL passed to urlopen
        called_url = mock_urlopen.call_args[0][0]
        self.assertNotIn("maxlat", called_url)
        self.assertNotIn("minlon", called_url)
        self.assertNotIn("maxlon", called_url)
        self.assertNotIn("minlat", called_url)


if __name__ == '__main__':
    unittest.main()



"""
This is a python version of the wave_signif.pro script in IDL SPEDAS.
It computes the significance levels for a wavelet transform
as described by Torrence & Compo (1998).

Reference: Torrence, C. and G. P. Compo, 1998: A Practical Guide to
           Wavelet Analysis. <I>Bull. Amer. Meteor. Soc.</I>, 79, 61-78.

See also: wavelet98.py
"""

import numpy as np
from scipy.stats import chi2

def wave_signif(Y, dt, scale, sigtest, **kwargs):
    """
    Compute significance levels for wavelet transforms.

    Parameters:
    -----------

    Y : float or array_like
        The time series, or, the VARIANCE of the time series.
        (If this is a single number, it is assumed to be the variance...)
    dt : float
        Amount of time between each Y value, i.e. the sampling time.
    scale : array_like
        The vector of scale indices, from previous call to WAVELET.
    sigtest : int {0, 1, 2}
        Type of significance test::

            If 0 (the default), then just do a regular chi-square test,
               i.e. Eqn (18) from Torrence & Compo.
            If 1, then do a "time-average" test, i.e. Eqn (23).
                     In this case, DOF should be set to NA, the number
                     of local wavelet spectra that were averaged together.
                     For the Global Wavelet Spectrum, this would be NA=N,
                     where N is the number of points in your time series.
            If 2, then do a "scale-average" test, i.e. Eqns (25)-(28).
                     In this case, DOF should be set to a
                     two-element vector [S1,S2], which gives the scale
                     range that was averaged together.
                     e.g. if one scale-averaged scales between 2 and 8,
                         then DOF=[2,8].

    kwargs : dict
        Dictionary, containing optional additional parameters::

            lag1: Lag-1 autocorrelation (default: 0.0)
            siglvl: Significance level (default: 0.95)
            mother: Mother wavelet ('MORLET', 'PAUL', 'DOG')
            param: Mother wavelet parameter
            dof: Degrees of freedom for significance tests
            gws: Global wavelet spectrum (optional)
            confidence: If True, return confidence intervals.

    Returns:
    --------
    signif : array_like
        Significance levels for the wavelet power spectrum.
    outputs : dict
        Dictionary containing optional output parameters::

            'Cdelta': Reconstruction factor
            'period': Vector of Fourier periods corresponding to scales
            'fft_theor': Theoretical red-noise spectrum as function of period
            'Savg': Scale average (for sigtest=2 only)
            'Smid': Scale midpoint (for sigtest=2 only)
    """

    # Default parameters
    lag1 = kwargs.get("lag1", 0.0)
    siglvl = kwargs.get("siglvl", 0.95)
    mother = kwargs.get("mother", "MORLET")
    param = kwargs.get("param", -1)
    dof = kwargs.get("dof", None)
    gws = kwargs.get("gws", None)
    confidence = kwargs.get("confidence", False)

    # Calculate variance of the input time series
    if np.isscalar(Y):
        variance = Y  # If Y is a scalar, assume it's the variance
    else:
        variance = np.var(Y, ddof=0)  # Use population variance like IDL MOMENT

    # Get scale parameters
    J = len(scale) - 1
    s0 = np.min(scale)
    if len(scale) > 1:
        dj = np.log(scale[1] / scale[0]) / np.log(2)
    else:
        dj = 0.25  # Default value

    # Fourier factor and empirical values for the chosen wavelet
    if mother.upper() == "MORLET":
        k0 = 6.0 if param == -1 else param
        fourier_factor = (4 * np.pi) / (k0 + np.sqrt(2 + k0**2))
        empir = [2.0, -1, -1, -1]
        if k0 == 6:
            empir[1:4] = [0.776, 2.32, 0.60]
    elif mother.upper() == "PAUL":
        m = 4 if param == -1 else param
        fourier_factor = 4 * np.pi / (2 * m + 1)
        empir = [2.0, -1, -1, -1]
        if m == 4:
            empir[1:4] = [1.132, 1.17, 1.5]
    elif mother.upper() == "DOG":
        m = 2 if param == -1 else param
        fourier_factor = 2 * np.pi * np.sqrt(2.0 / (2 * m + 1))
        empir = [1.0, -1, -1, -1]
        if m == 2:
            empir[1:4] = [3.541, 1.43, 1.4]
        elif m == 6:
            empir[1:4] = [1.966, 1.37, 0.97]
    else:
        raise ValueError(f"Unsupported mother wavelet: {mother}")

    # Equivalent Fourier period
    period = scale * fourier_factor

    # Empirical constants
    dofmin = empir[0]  # Minimum degrees of freedom
    Cdelta = empir[1]  # Reconstruction factor
    gamma = empir[2]  # Time-decorrelation factor
    dj0 = empir[3]  # Scale-decorrelation factor

    # Compute the red-noise spectrum [Eqn(16)]
    freq = dt / period  # normalized frequency
    fft_theor = (1 - lag1**2) / (1 - 2 * lag1 * np.cos(freq * 2 * np.pi) + lag1**2)
    fft_theor = variance * fft_theor  # include time-series variance

    # Use global wavelet spectrum if provided
    if gws is not None and len(gws) == len(scale):
        fft_theor = gws

    signif = fft_theor.copy()

    # Initialize output dictionary
    outputs = {
        "Cdelta": Cdelta,
        "period": period,
        "fft_theor": fft_theor,
        "Savg": None,
        "Smid": None,
    }

    # Significance levels [Sec.4]
    if sigtest == 0:  # Local significance (no smoothing) [Eqn(18)]
        dof_local = dofmin
        signif = fft_theor * chi2.ppf(siglvl, dof_local) / dof_local
        if confidence:
            sig = (1.0 - siglvl) / 2.0
            chisqr = dof_local / np.array(
                [chi2.ppf(sig, dof_local), chi2.ppf(1.0 - sig, dof_local)]
            )
            signif = fft_theor[:, None] * chisqr

    elif sigtest == 1:  # Global significance (time-averaged) [Sec.5a]
        if gamma == -1:
            raise ValueError(
                f"Gamma (decorrelation factor) not defined for {mother} with param={param}"
            )

        if dof is None:
            dof_local = np.full(J + 1, dofmin)
        else:
            if np.isscalar(dof):
                dof_local = np.full(J + 1, dof)
            else:
                dof_local = np.array(dof)

        dof_local = np.maximum(dof_local, 1)
        dof_local = dofmin * np.sqrt(
            1 + (dof_local * dt / gamma / scale) ** 2
        )  # [Eqn(23)]
        dof_local = np.maximum(dof_local, dofmin)  # minimum DOF is dofmin

        if not confidence:
            signif = np.zeros_like(fft_theor)
            for i in range(J + 1):
                chisqr = chi2.ppf(siglvl, dof_local[i]) / dof_local[i]
                signif[i] = fft_theor[i] * chisqr
        else:
            signif = np.zeros((J + 1, 2))
            sig = (1.0 - siglvl) / 2.0
            for i in range(J + 1):
                chisqr = dof_local[i] / np.array(
                    [chi2.ppf(sig, dof_local[i]), chi2.ppf(1.0 - sig, dof_local[i])]
                )
                signif[i, :] = fft_theor[i] * chisqr

    elif sigtest == 2:  # Scale-average significance [Sec.5b]
        if dof is None or len(dof) != 2:
            raise ValueError("DOF must be set to [S1,S2], the range of scale-averages")
        if Cdelta == -1:
            raise ValueError(
                f"Cdelta & dj0 not defined for {mother} with param={param}"
            )

        s1, s2 = dof
        avg_idx = np.where((scale >= s1) & (scale <= s2))[0]
        if len(avg_idx) < 1:
            raise ValueError(f"No valid scales between {s1} and {s2}")

        s1 = np.min(scale[avg_idx])
        s2 = np.max(scale[avg_idx])
        Savg = 1.0 / np.sum(1.0 / scale[avg_idx])  # [Eqn(25)]
        Smid = np.exp((np.log(s1) + np.log(s2)) / 2.0)  # power-of-two midpoint
        navg = len(avg_idx)
        dof_eff = (dofmin * navg * Savg / Smid) * np.sqrt(
            1 + (navg * dj / dj0) ** 2
        )  # [Eqn(28)]
        fft_theor_avg = Savg * np.sum(fft_theor[avg_idx] / scale[avg_idx])  # [Eqn(27)]
        chisqr = chi2.ppf(siglvl, dof_eff) / dof_eff

        # Store scale-average outputs
        outputs["Savg"] = Savg
        outputs["Smid"] = Smid

        if confidence:
            sig = (1.0 - siglvl) / 2.0
            chisqr = dof_eff / np.array(
                [chi2.ppf(sig, dof_eff), chi2.ppf(1.0 - sig, dof_eff)]
            )

        signif = (dj * dt / Cdelta / Savg) * fft_theor_avg * chisqr  # [Eqn(26)]

    else:
        raise ValueError(f"Unsupported sigtest value: {sigtest}")

    return signif, outputs



"""
Displays location of source files, one line of documentation and the function name based on the request
"""

import importlib
import inspect
import itertools
import pkgutil
import sys
import io
from fnmatch import fnmatchcase

import pyspedas
import functools


def libs(function_name, package=None):
    """
    Search for a specified function by exact name, substring, or wildcard within a given package and its submodules.

    Parameters
    ----------
    function_name (str):
        The name or partial name of the function to search for. If "*" or "?" are found in the function name, a case-insensitive wildcard match is performed, otherwise
        function_name is treated as a substring to match.

    package (module, optional): The Python package in which to search for the function.
        Default is the pyspedas package. This should be a Python module object.

    Note
    ----

    All submodules of pyspedas are imported during the search. The package option is
    simply narrows the search.
    The function specifically searches for functions, not classes or other objects.
    If multiple functions with the same name exist in different modules within the package, it will list them all.
    The function handles ImportError exceptions by printing an error message and
    continuing the search.

    Example
    -------

    >>> from pyspedas import libs
    >>> libs('fgm')
    >>> libs('fgm', package=pyspedas.projects.mms)

    """

    # Gate for no function_name
    if not function_name:
        return

    # Using separate functions for the wildcard and substring matching eliminates a test and branch in the inner loop

    def list_functions_substring(module, root, search_string, pacakge_obj):
        full_module_name = module.__name__
        for name, obj in inspect.getmembers(module):
            if inspect.isfunction(obj) and (search_string in name) and (pacakge_obj.__name__ in obj.__module__):
                full_name = full_module_name + '.' + name
                source_file = inspect.getsourcefile(obj)
                doc = inspect.getdoc(obj)
                first_line_of_doc = doc.split('\n')[0] if doc else "No documentation"
                print(f"Function: {full_name}\nLocation: {source_file}\nDocumentation: {first_line_of_doc}\n")
            elif isinstance(obj, functools.partial) and (search_string in name):
                original_func = obj.func
                full_name = full_module_name + '.' + name
                source_file = inspect.getsourcefile(original_func)
                doc = inspect.getdoc(original_func)
                first_line_of_doc = doc.split('\n')[0] if doc else "No documentation"
                print(f"Partial Function: {full_name}\nLocation: {source_file}\nDocumentation: {first_line_of_doc}\n")

    def list_functions_wildcard(module, root, wildcard_pattern, pacakge_obj):
        full_module_name = module.__name__
        for name, obj in inspect.getmembers(module):
            if inspect.isfunction(obj) and fnmatchcase(name.lower(), wildcard_pattern) and (pacakge_obj.__name__ in obj.__module__):
                full_name = full_module_name + '.' + name
                source_file = inspect.getsourcefile(obj)
                doc = inspect.getdoc(obj)
                first_line_of_doc = doc.split('\n')[0] if doc else "No documentation"
                print(f"Function: {full_name}\nLocation: {source_file}\nDocumentation: {first_line_of_doc}\n")
            elif isinstance(obj, functools.partial) and fnmatchcase(name.lower(), wildcard_pattern):
                original_func = obj.func
                full_name = full_module_name + '.' + name
                source_file = inspect.getsourcefile(original_func)
                doc = inspect.getdoc(original_func)
                first_line_of_doc = doc.split('\n')[0] if doc else "No documentation"
                print(f"Partial Function: {full_name}\nLocation: {source_file}\nDocumentation: {first_line_of_doc}\n")

    def traverse_modules(package, function_name, package_obj):
        # Add the module itself
        walk_packages_iterator = pkgutil.walk_packages(path=package.__path__, prefix=package.__name__ + '.')
        combined_iterator = itertools.chain(((None, package.__name__, True),), walk_packages_iterator)

        # Check for wildcard characters
        if '*' in function_name or '?' in function_name:
            wildcard=True
            # There is no 'fnmatchnocase', so lowercase the search pattern and function names before comparing
            # We'll add implicit leading and trailing '*', so any substring match will appear in the list
            wildcard_pattern = '*' + function_name.lower() + '*'
        else:
            wildcard=False
            wildcard_pattern=''

        for _, modname, ispkg in combined_iterator:
            if ispkg and 'qtplotter' not in modname.lower():
                # Save the current stdout so that we can restore it later
                original_stdout = sys.stdout

                # Redirect stdout to a dummy stream
                sys.stdout = io.StringIO()

                try:
                    module = importlib.import_module(modname)
                    if not package_obj:
                        package_obj = package

                    # Restore the original stdout
                    sys.stdout = original_stdout
                    if not wildcard:
                        list_functions_substring(module, package, function_name, package_obj)
                    else:
                        list_functions_wildcard(module, package, wildcard_pattern, package_obj)
                except ImportError as e:
                    # Restore the original stdout
                    sys.stdout = original_stdout
                    print(f"Error importing module {modname}: {e}")
                finally:
                    # Restore the original stdout
                    sys.stdout = original_stdout

    for module in [pyspedas]:
        if not package or module.__name__ in package.__name__:
            traverse_modules(module, function_name, package)



"""
This module contains routines for loading MMS data

"""

from functools import wraps

from .mec_ascii.state import mms_load_state
from .mec_ascii.tetrahedron_qf import mms_load_tetrahedron_qf
from .fgm_tools.fgm import mms_load_fgm
from .hpca_tools.hpca import mms_load_hpca
from .fpi_tools.fpi import mms_load_fpi
from .fpi_tools.mms_fpi_ang_ang import mms_fpi_ang_ang
from .scm_tools.scm import mms_load_scm
from .mec_tools.mec import mms_load_mec
from .feeps_tools.feeps import mms_load_feeps
from .eis_tools.eis import mms_load_eis
from .edi_tools.edi import mms_load_edi
from .edp_tools.edp import mms_load_edp
from .dsp_tools.dsp import mms_load_dsp
from .aspoc_tools.aspoc import mms_load_aspoc
from .fsm_tools.fsm import mms_load_fsm
from .fgm_tools.mms_curl import mms_curl
from .fgm_tools.mms_lingradest import mms_lingradest
from .spd_mms_load_bss import spd_mms_load_bss
from .cotrans.mms_cotrans_lmn import mms_cotrans_lmn
from .cotrans.mms_cotrans_qrotate import mms_cotrans_qrotate
from .cotrans.mms_cotrans_qtransformer import mms_cotrans_qtransformer
from .cotrans.mms_qcotrans import mms_qcotrans
from .particles import mms_part_getspec, mms_part_slice2d
from .mms_orbit_plot import mms_orbit_plot
from .mms_load_brst_segments import mms_load_brst_segments
from pyspedas.projects.mms.deprecated.mms_load_fast_segments import mms_load_fast_segments
from .mms_load_sroi_segments import mms_load_sroi_segments
from .mms_events import mms_brst_events

'''
    the following wrappers allow users to import the load routines using 
    the syntax: 
    
            >>> from pyspedas.projects.mms import fgm
            >>> fgm_data = fgm(...)

        and/or

            >>> import pyspedas
            >>> fgm_data = pyspedas.projects.mms.fgm(...)
'''

# Some of these wrappers shadow MMS module names in a way that seems too fragile -- adding an apparently unrelated
# import in the "wrong" place can start triggering errors like "module fgm is not callable".  It might be better and
# more robust to rename the instrument directories so they don't conflict with the wrapper names.  Also,
# perhaps these wrapper names should be the "real" names, with the longer names mms_load_fgm, mms_load_state, etc.
# as wrappers or aliases rather than the other way around.   At least in PyCharm, hovering over a call to fgm() to
# see the argument list only shows the *args and **kwargs parameters, which is not terribly useful.
# JWL 2024-03-29

@wraps(spd_mms_load_bss)
def bss(*args, **kwargs):
    return spd_mms_load_bss(*args, **kwargs)

@wraps(mms_load_state)
def state(*args, **kwargs):
    return mms_load_state(*args, **kwargs)

@wraps(mms_load_tetrahedron_qf)
def tetrahedron_qf(*args, **kwargs):
    return mms_load_tetrahedron_qf(*args, **kwargs)

@wraps(mms_load_fgm)
def fgm(*args, **kwargs):
    return mms_load_fgm(*args, **kwargs)

@wraps(mms_load_scm)
def scm(*args, **kwargs):
    return mms_load_scm(*args, **kwargs)

@wraps(mms_load_fsm)
def fsm(*args, **kwargs):
    return mms_load_fsm(*args, **kwargs)

@wraps(mms_load_edp)
def edp(*args, **kwargs):
    return mms_load_edp(*args, **kwargs)

@wraps(mms_load_edi)
def edi(*args, **kwargs):
    return mms_load_edi(*args, **kwargs)

@wraps(mms_load_fpi)
def fpi(*args, **kwargs):
    return mms_load_fpi(*args, **kwargs)

@wraps(mms_load_hpca)
def hpca(*args, **kwargs):
    return mms_load_hpca(*args, **kwargs)

@wraps(mms_load_eis)
def eis(*args, **kwargs):
    return mms_load_eis(*args, **kwargs)

@wraps(mms_load_feeps)
def feeps(*args, **kwargs):
    return mms_load_feeps(*args, **kwargs)

@wraps(mms_load_aspoc)
def aspoc(*args, **kwargs):
    return mms_load_aspoc(*args, **kwargs)

@wraps(mms_load_mec)
def mec(*args, **kwargs):
    return mms_load_mec(*args, **kwargs)

@wraps(mms_load_dsp)
def dsp(*args, **kwargs):
    return mms_load_dsp(*args, **kwargs)

@wraps(mms_curl)
def curlometer(*args, **kwargs):
    return mms_curl(*args, **kwargs)

@wraps(mms_lingradest)
def lingradest(*args, **kwargs):
    return mms_lingradest(*args, **kwargs)



import numpy as np
from pyspedas.tplot_tools import get_data, options, store_data, tplot_names

from .cart_trans_matrix_make import cart_trans_matrix_make
from .erg_interpolate_att import erg_interpolate_att


def sga2sgi(name_in=None,
            name_out=None,
            SGI2SGA=False,
            noload=False):
    """
    This transform a time series data between the SGA and SGI coordinate systems.

    Parameters:

        name_in : str
            input tplot variable to be transformed

        name_out : str
            Name of the tplot variable in which the transformed data is stored

        SGI2SGA : bool
              Set to transform data from SGI to SGA. If not set, it transforms data from SGA to SGI.

    Returns:
        None

    """
    if (name_in is None) or (name_in not in tplot_names(quiet=True)):
        print('Input of Tplot name is undifiend')
        return

    if name_out is None:
        print('Tplot name for output is undifiend')
        name_out = 'result_of_sga2sgi'

    get_data_vars = get_data(name_in)
    dl_in = get_data(name_in, metadata=True)
    time_array = get_data_vars[0]
    time_length = time_array.shape[0]
    dat = get_data_vars[1]

    # Get the SGA and SGI axes by interpolating the attitude data
    interpolated_values = erg_interpolate_att(name_in, noload=noload)
    sgix = interpolated_values['sgix_j2000']['y']
    sgiy = interpolated_values['sgiy_j2000']['y']
    sgiz = interpolated_values['sgiz_j2000']['y']
    sgax = interpolated_values['sgax_j2000']['y']
    sgay = interpolated_values['sgay_j2000']['y']
    sgaz = interpolated_values['sgaz_j2000']['y']

    if not SGI2SGA:
        print('SGA --> SGI')
        coord_out = 'sgi'

        # Transform SGI-X,Y,Z axis unit vectors in J2000 to those in SGA
        mat = cart_trans_matrix_make(sgax, sgay, sgaz)
        sgix_in_sga = np.einsum("ijk,ik->ij", mat, sgix)
        sgiy_in_sga = np.einsum("ijk,ik->ij", mat, sgiy)
        sgiz_in_sga = np.einsum("ijk,ik->ij", mat, sgiz)

        # Now transform the given vector in SGA to those in SGI
        mat = cart_trans_matrix_make(sgix_in_sga, sgiy_in_sga, sgiz_in_sga)
        dat_new = np.einsum("ijk,ik->ij", mat, dat)

    else:
        print('SGI --> SGA')
        coord_out = 'sga'

        # Transform SGA-X,Y,Z axis unit vectors in J2000 to those in SGI
        mat = cart_trans_matrix_make(sgix, sgiy, sgiz)
        sgax_in_sgi = np.einsum("ijk,ik->ij", mat, sgax)
        sgay_in_sgi = np.einsum("ijk,ik->ij", mat, sgay)
        sgaz_in_sgi = np.einsum("ijk,ik->ij", mat, sgaz)

        # Now transform the given vector in SGI to those in SGA
        mat = cart_trans_matrix_make(sgax_in_sgi, sgay_in_sgi, sgaz_in_sgi)
        dat_new = np.einsum("ijk,ik->ij", mat, dat)

    # Store the converted data in a tplot variable
    store_data(name_out, data={'x': time_array, 'y': dat_new}, attr_dict=dl_in)
    options(name_out, 'ytitle', '\n'.join(name_out.split('_')))




def spd_pgs_limit_range(data, phi=None, theta=None, energy=None):
    """
    Applies phi, theta, and energy limits to data structure(s) by
    turning off the corresponding bin flags.

    Parameters
    ----------
        data: dict
            Particle data structure

        phi: np.ndarray
            Minimum and maximum values for phi

        theta: np.ndarray
            Minimum and maximum values for theta

        energy: np.ndarray
            Minimum and maximum values for energy

    Returns
    -------
    dict
        Data structure with limits applied (to the bins array)
    """

    # if no limits are set, return the input data
    if energy is None and theta is None and phi is None:
        return data

    # apply the phi limits
    if phi is not None:
        # get min/max phi values for all bins
        phi_min = data['phi'] - 0.5*data['dphi']
        phi_max = data['phi'] + 0.5*data['dphi'] % 360

        # wrap negative values
        phi_min[phi_min < 0.0] += 360

        # the code below and the phi spectrogram code 
        # assume maximums at 360 are not wrapped to 0
        phi_max[phi_max == 0.0] = 360.0

        # find which bins were wrapped back into [0, 360]
        wrapped = phi_min > phi_max

        # determine which bins intersect the specified range
        if phi[0] > phi[1]:
            in_range = phi_min < phi[1] or phi_max > phi[0] or wrapped
        else:
            in_range = ((phi_min < phi[1]) & (phi_max > phi[0])) | (wrapped & ((phi_min < phi[1]) | (phi_max > phi[0])))

        data['bins'][in_range == False] = 0

    # apply the theta limits
    if theta is not None:
        lower_theta = min(theta)
        upper_theta = max(theta)

        # get min/max angle theta values for all bins
        theta_min = data['theta'] - 0.5*data['dtheta']
        theta_max = data['theta'] + 0.5*data['dtheta']
        
        in_range = (theta_min < upper_theta) & (theta_max > lower_theta)
        data['bins'][in_range == False] = 0

    # apply the energy limits
    if energy is not None:
        data['bins'][data['energy'] < energy[0]] = 0
        data['bins'][data['energy'] > energy[1]] = 0

    return data



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot
from typing import List, Union
from .config import CONFIG

def load(trange:List[str]=['2018-11-5', '2018-11-6'],
         probe:Union[str,List[str]]='1',
         instrument:str='lad',
         datatype:str='',
         prefix:str='',
         suffix:str='',
         get_support_data:bool=False,
         varformat:str=None,
         varnames:List[str]=[],
         downloadonly:bool=False,
         force_download:bool=False,
         notplot:bool=False,
         no_update:bool=False,
         time_clip:bool=False) -> List[str]:
    """Load data from the TWINS mission
        Parameters
        ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2018-11-5', '2018-11-6']

        probe: str or list of str
            Probe to load. Valid options: '1', '2'
            Default: '1'

        instrument : str
            The instrument to load. Valid options: 'lad', 'imager', 'ephemeris'
            Default: 'lad'

        datatype: str
            Data type; Valid options: ''
            Default: ''

        prefix: str
            The tplot variable names will be given this prefix.
            Default: ''

        suffix: str
            The tplot variable names will be given this suffix.
            Default: ''

        get_support_data: bool
            If True, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: '' (all variables loaded)

        varnames: list of str
            List of variable names to load
            Default: [] (all variables will be loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        force_download: bool
            If True, download the file even if the local version is newer
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

    Returns
    --------
        list of str
            List of tplot variables created.

    Notes
    -----
    This function loads TWINS data; this function is not meant 
    to be called directly; instead, see the wrappers:
        pyspedas.projects.twins.lad
        pyspedas.projects.twins.ephemeris
        pyspedas.projects.twins.imager

    """

    if prefix is None:
        prefix = ''
    if suffix is None:
        suffix = ''

    if not isinstance(probe, list):
        probe = [probe]

    probe = [str(prb) for prb in probe]

    out_files = []

    for prb in probe:
        if instrument == 'lad':
            pathformat = 'twins'+prb+'/'+instrument+'/%Y/twins'+prb+'_l1_lad_%Y%m%d_v??.cdf'
        elif instrument == 'imager':
            pathformat = 'twins'+prb+'/'+instrument+'/%Y/twins'+prb+'_l1_imager_%Y%m%d??_v??.cdf'
        elif instrument == 'ephemeris':
            pathformat = 'twins'+prb+'/'+instrument+'/'+datatype+'/%Y/twins'+prb+'_'+datatype+'_def_%Y%m%d_v??.cdf'

        # find the full remote path names using the trange
        remote_names = dailynames(file_format=pathformat, trange=trange)

        files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update, force_download=force_download)
        if files is not None:
            for file in files:
                out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, prefix=prefix, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars


import numpy as np
from pyspedas.tplot_tools import store_data, options
from pyspedas.tplot_tools import time_double
from .load_csv_file import load_csv_file


# This routine was originally in akebono/__init__.py.  If you need to see the history of this routine before
# it was moved to its own file, please check the history for __init__.py.

def rdm_postprocessing(files, prefix='', suffix=''):
    """
    Load the RDM ASCII files into tplot variables

    Parameters
    ----------
    files: list of str
        List of filenames to convert to tplot variables
    prefix: str
        Prefix to be added to tplot variables created. Default: ''
    suffix: str
        Suffix to be added to tplot variables created. Default: ''

    Returns
    -------
    list of str
    List of tplot variables created.
    """

    if prefix is None:
        user_prefix = ''
    else:
        user_prefix = prefix

    if suffix is None:
        suffix = ''

    try:
        data = load_csv_file(files)
    except UnicodeDecodeError:
        data = load_csv_file(files, gz=True)

    values = data.to_numpy()
    unix_times = time_double([ymd + '/' + hms for ymd, hms in zip(values[:, 0], values[:, 1])])

    L = np.float64(values[:, 2])
    INV = np.float64(values[:, 3])
    FMLAT = np.float64(values[:, 4])
    MLAT = np.float64(values[:, 5])
    MLT = np.float64(values[:, 6])
    ALT = np.float64(values[:, 7])
    GLAT = np.float64(values[:, 8])
    GLON = np.float64(values[:, 9])
    RDM_E3 = np.float64(values[:, 10])
    Energy = np.zeros(len(RDM_E3))
    Energy[:] = 2.5

    prefix_project = 'akb_'
    prefix_descriptor = 'rdm_'
    prefix = user_prefix + prefix_project + prefix_descriptor

    store_data(user_prefix + prefix_project+'L' + suffix, data={'x': unix_times, 'y': L})
    store_data(user_prefix + prefix_project+'INV' + suffix, data={'x': unix_times, 'y': INV})
    store_data(user_prefix + prefix_project+'FMLAT' + suffix, data={'x': unix_times, 'y': FMLAT})
    store_data(user_prefix + prefix_project+'MLAT' + suffix, data={'x': unix_times, 'y': MLAT})
    store_data(user_prefix + prefix_project+'MLT' + suffix, data={'x': unix_times, 'y': MLT})
    store_data(user_prefix + prefix_project+'ALT' + suffix, data={'x': unix_times, 'y': ALT})
    store_data(user_prefix + prefix_project+'GLAT' + suffix, data={'x': unix_times, 'y': GLAT})
    store_data(user_prefix + prefix_project+'GLON' + suffix, data={'x': unix_times, 'y': GLON})
    store_data(prefix+'FEIO' + suffix, data={'x': unix_times, 'y': RDM_E3})
    store_data(prefix+'FEIO_Energy' + suffix, data={'x': unix_times, 'y': Energy})

    # Not sure why this is here...FEIO is a scalar at each time step
    options(prefix+'FEIO' + suffix, 'spec', True)

    options(user_prefix + prefix_project+'L' + suffix, 'ytitle', 'L-value')
    options(user_prefix + prefix_project+'INV' + suffix, 'ytitle', 'Invariant Latitude [deg]')
    options(user_prefix + prefix_project+'FMLAT' + suffix, 'ytitle', 'Footprint Latitude [deg]')
    options(user_prefix + prefix_project+'MLAT' + suffix, 'ytitle', 'Magnetic Latitude [deg]')
    options(user_prefix + prefix_project+'MLT' + suffix, 'ytitle', 'Magnetic Local Time [hour]')
    options(user_prefix + prefix_project+'ALT' + suffix, 'ytitle', 'Altitude [km]')
    options(user_prefix + prefix_project+'GLAT' + suffix, 'ytitle', 'Geographic Latitude [deg]')
    options(user_prefix + prefix_project+'GLON' + suffix, 'ytitle', 'Geographic Longitude [deg]')
    options(prefix+'FEIO' + suffix, 'ytitle', 'Omni-directional Integral Electron Flux')
    options(prefix+'FEIO' + suffix, 'ysubtitle', '[/cm^22 sec str]')
    options(prefix+'FEIO_Energy' + suffix, 'ytitle', 'Elctron energy [MeV]')

    return [user_prefix + prefix_project+'L' + suffix,
            user_prefix + prefix_project+'INV' + suffix,
            user_prefix + prefix_project+'FMLAT' + suffix,
            user_prefix + prefix_project+'MLAT' + suffix,
            user_prefix + prefix_project+'MLT' + suffix,
            user_prefix + prefix_project+'ALT' + suffix,
            user_prefix + prefix_project+'GLAT' + suffix,
            user_prefix + prefix_project+'GLON' + suffix,
            prefix+'FEIO' + suffix,
            prefix+'FEIO_Energy' + suffix]





"""
Convert Cartesian coordinates to spherical coordinates.

This function is similar to cart_to_sphere.pro in IDL SPEDAS.
"""

import numpy as np


def cart_to_sphere(x, y, z):
    """
    Convert Cartesian coordinates to spherical coordinates.

    Parameters:
    -----------
    x : float
        X coordinate in Cartesian system.
    y : float
        Y coordinate in Cartesian system.
    z : float
        Z coordinate in Cartesian system.

    Returns:
    --------
    r : float
        Radial distance from the origin.
    theta : float
        Altitude angle (not co-latitude!) from XY plane to point, in degrees. It ranges from -90 to 90 degrees
    phi : float
        Azimuthal angle (angle in the xy-plane from the positive x-axis to the point) in radians. It ranges from 0 to 360 degrees.

    Notes:
    ------
    - Uses the following equations for conversion:
        r = sqrt(x^2 + y^2 + z^2)
        theta = arcsin(z/r)
        phi = 2*pi - arccos(x/sqrt(x^2 + y^2)) if y < 0, otherwise arccos(x/sqrt(x^2 + y^2))
    """

    x, y, z = np.array(x), np.array(y), np.array(z)
    r = np.sqrt(x**2 + y**2 + z**2)

    # Calculate phi
    phi = np.arctan2(y,x)

    # Ensure range is in [0,2*pi]
    corr = np.where(phi < 0.0,2.0*np.pi, 0)
    phi += corr

    # Calculate theta
    theta = np.arcsin(z/r)

    return r, np.rad2deg(theta), np.rad2deg(phi)



import os
import unittest

import numpy as np
from numpy.testing import assert_allclose
from geopack import geopack

import pyspedas
from pyspedas import (
    del_data,
    tplot,
    subtract,
    tlimit,
    time_double,
    tplot_restore,
    tinterpol,
    tvectot,
    join_vec,
    store_data,
    get_data,
    tkm2re,
)
from pyspedas.geopack import tt89, tt96, tt01, tts04
from pyspedas.geopack.get_tsy_params import get_tsy_params
from pyspedas.utilities.config_testing import TESTING_CONFIG, test_data_download_file

# Whether to display plots during testing
global_display = TESTING_CONFIG["global_display"]
# Directory to save testing output files
output_dir = TESTING_CONFIG["local_testing_dir"]
# Ensure output directory exists
geopack_dir = "geopack"
save_dir = os.path.join(output_dir, geopack_dir)
if not os.path.exists(save_dir):
    os.makedirs(save_dir)
# Directory with IDL SPEDAS validation files
validation_dir = TESTING_CONFIG["remote_validation_dir"]


trange = ["2015-10-16", "2015-10-17"]


def get_params(model, g_variables=None):
    support_trange = [
        time_double(trange[0]) - 60 * 60 * 24,
        time_double(trange[1]) + 60 * 60 * 24,
    ]
    pyspedas.projects.kyoto.dst(trange=support_trange)
    pyspedas.projects.omni.data(trange=trange)
    join_vec(["BX_GSE", "BY_GSM", "BZ_GSM"])
    if model == "t01" and g_variables is None:
        g_variables = [6.0, 10.0]
    else:
        if g_variables is not None:
            if not isinstance(g_variables, str) and not isinstance(
                g_variables, np.ndarray
            ):
                g_variables = None
    return get_tsy_params(
        "kyoto_dst",
        "BX_GSE-BY_GSM-BZ_GSM_joined",
        "proton_density",
        "flow_speed",
        model,
        pressure_tvar="Pressure",
        g_variables=g_variables,
        speed=True,
    )


class LoadGeopackIdlValidationTestCases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """
        IDL Data has to be downloaded to perform these tests
        The SPEDAS script that creates the file: general/tools/python_validate/geopack_validate.pro
        """
        # Testing time range
        cls.t = ["2008-03-23", "2008-03-28"]

        # Testing tolerance
        cls.tol = 1e-10

        # Download tplot files
        filename = test_data_download_file(
            validation_dir,
            "analysis_tools",
            "geopack_idl_validate_multi_gp14.tplot",
            save_dir,
        )

        del_data("*")
        tplot_restore(filename)

        # Input parameters
        cls.iopt = 3
        cls.kp = 2.0
        cls.pdyn = 2.0
        cls.dsti = -30.0
        cls.yimf = 0.0
        cls.zimf = -5.0
        cls.g1 = 6.0
        cls.g2 = 10.0
        cls.w1 = 8.0
        cls.w2 = 5.0
        cls.w3 = 9.5
        cls.w4 = 30.0
        cls.w5 = 18.5
        cls.w6 = 60.0

    def model_circ_data_tt01(self, year: str):
        tv_pos = get_data(f"circle_magpoles_5re_{year}_km")
        t1 = tv_pos.times[0]
        t2 = tv_pos.times[-1]
        params = np.zeros([2, 10])
        params[:, 0] = self.pdyn
        params[:, 1] = self.dsti
        params[:, 2] = self.yimf
        params[:, 3] = self.zimf
        params[:, 4] = self.g1
        params[:, 5] = self.g2
        store_data("parmod", data={"x": [t1, t2], "y": params})
        tinterpol(
            "parmod",
            f"circle_magpoles_5re_{year}_km",
            method="nearest",
            newname="parmod_interp",
        )
        tt01(f"circle_magpoles_5re_{year}_km", parmod="parmod_interp")
        subtract(
            f"tst5re_{year}_bt01",
            f"circle_magpoles_5re_{year}_km_bt01",
            f"bt01_{year}_diff",
        )
        # tkm2re('tha_state_pos_gsm')
        # tvectot('tha_state_pos_gsm_re',join_component=True)

        var_names = [
            f"tst5re_{year}_bt01",
            f"circle_magpoles_5re_{year}_km_bt01",
            f"bt01_{year}_diff",
        ]
        local_png = os.path.join(save_dir, f"circ_t01_{year}_diffs.png")
        tplot(var_names, display=global_display, save_png=local_png)

    def model_circ_data_tts04(self, year: str):
        tv_pos = get_data(f"circle_magpoles_5re_{year}_km")
        t1 = tv_pos.times[0]
        t2 = tv_pos.times[-1]
        params = np.zeros([2, 10])
        params[:, 0] = self.pdyn
        params[:, 1] = self.dsti
        params[:, 2] = self.yimf
        params[:, 3] = self.zimf
        params[:, 4] = self.w1
        params[:, 5] = self.w2
        params[:, 6] = self.w3
        params[:, 7] = self.w4
        params[:, 8] = self.w5
        params[:, 9] = self.w6
        store_data("parmod", data={"x": [t1, t2], "y": params})
        tinterpol(
            "parmod",
            f"circle_magpoles_5re_{year}_km",
            method="nearest",
            newname="parmod_interp",
        )
        tts04(f"circle_magpoles_5re_{year}_km", parmod="parmod_interp")
        py_b = get_data(f"circle_magpoles_5re_{year}_km_bts04")
        idl_b = get_data(f"tst5re_{year}_bts04")
        subtract(
            f"tst5re_{year}_bts04",
            f"circle_magpoles_5re_{year}_km_bts04",
            f"bts04_{year}_diff",
        )
        var_names = [
            f"tst5re_{year}_bts04",
            f"circle_magpoles_5re_{year}_km_bts04",
            f"bts04_{year}_diff",
        ]
        local_png = os.path.join(save_dir, f"circ_ts04_{year}_diffs.png")
        tplot(var_names, display=global_display, save_png=local_png)

        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def model_circ_data_tt96(self, year: str):
        tv_pos = get_data(f"circle_magpoles_5re_{year}_km")
        t1 = tv_pos.times[0]
        t2 = tv_pos.times[-1]
        params = np.zeros([2, 10])
        params[:, 0] = self.pdyn
        params[:, 1] = self.dsti
        params[:, 2] = self.yimf
        params[:, 3] = self.zimf
        store_data("parmod", data={"x": [t1, t2], "y": params})
        tinterpol(
            "parmod",
            f"circle_magpoles_5re_{year}",
            method="nearest",
            newname="parmod_interp",
        )
        tt96(f"circle_magpoles_5re_{year}_km", parmod="parmod_interp")
        # tkm2re('tha_state_pos_gsm')
        # tvectot('tha_state_pos_gsm_re',join_component=True)
        subtract(
            f"tst5re_{year}_bt96",
            f"circle_magpoles_5re_{year}_km_bt96",
            f"bt96_{year}_diff",
        )
        var_names = [
            f"tst5re_{year}_bt96",
            f"circle_magpoles_5re_{year}_km_bt96",
            f"bt96_{year}_diff",
        ]
        local_png = os.path.join(save_dir, f"circ_t96_{year}_diffs.png")
        tplot(var_names, display=global_display, save_png=local_png)
        # tlimit(['2007-03-23/15:00','2007-03-23/17:00'])
        # tplot(['bt96','tha_state_pos_gsm_bt96','bt96_diff','tha_state_pos_gsm_re_tot'], display=display)
        tlimit(full=True)

    def test_tilt(self):
        tilt_times, tilt_vals = get_data("bt89_tilt")
        py_tilt_vals = np.zeros(len(tilt_times))
        for idx, t in enumerate(tilt_times):
            py_tilt_vals[idx] = np.degrees(geopack.recalc(t))
        store_data("py_tilt", data={"x": tilt_times, "y": py_tilt_vals})
        idl_tilt = get_data("bt89_tilt")
        py_tilt = get_data("py_tilt")
        assert_allclose(idl_tilt.y, py_tilt.y, atol=0.0003)

    def test_igrf(self):
        tt89("tha_state_pos_gsm", igrf_only=True)
        py_b = get_data("tha_state_pos_gsm_bt89")
        idl_b = get_data("bt89_igrf")
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=1.0)

    def test_tt89(self):
        tt89("tha_state_pos_gsm")
        py_b = get_data("tha_state_pos_gsm_bt89")
        idl_b = get_data("bt89")
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_tt96(self):
        tv_pos = get_data("tha_state_pos_gsm")
        t1 = tv_pos.times[0]
        t2 = tv_pos.times[-1]
        params = np.zeros([2, 10])
        params[:, 0] = self.pdyn
        params[:, 1] = self.dsti
        params[:, 2] = self.yimf
        params[:, 3] = self.zimf
        store_data("parmod", data={"x": [t1, t2], "y": params})
        tinterpol(
            "parmod", "tha_state_pos_gsm", method="nearest", newname="parmod_interp"
        )
        tt96("tha_state_pos_gsm", parmod="parmod_interp")
        tkm2re("tha_state_pos_gsm")
        tvectot("tha_state_pos_gsm_re", join_component=True)
        py_b = get_data("tha_state_pos_gsm_bt96")
        idl_b = get_data("bt96")
        subtract("bt96", "tha_state_pos_gsm_bt96", "bt96_diff")
        var_names = [
            "bt96",
            "tha_state_pos_gsm_bt96",
            "bt96_diff",
            "tha_state_pos_gsm_re_tot",
        ]
        local_png = os.path.join(save_dir, "t96_diffs.png")
        tplot(var_names, display=global_display, save_png=local_png)

        tlimit(["2007-03-23/15:00", "2007-03-23/17:00"])
        local_png = os.path.join(save_dir, "t96_diffs1.png")
        tplot(var_names, display=global_display, save_png=local_png)

        tlimit(full=True)
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_tt96_circ_2026(self):
        self.model_circ_data_tt96("2026")
        py_b = get_data("circle_magpoles_5re_2026_km_bt96")
        idl_b = get_data("tst5re_2026_bt96")
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_tt96_circ_2024(self):
        self.model_circ_data_tt96("2024")
        py_b = get_data("circle_magpoles_5re_2024_km_bt96")
        idl_b = get_data("tst5re_2024_bt96")
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_tt96_circ_2019(self):
        self.model_circ_data_tt96("2019")
        py_b = get_data("circle_magpoles_5re_2019_km_bt96")
        idl_b = get_data("tst5re_2019_bt96")
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_tt96_circ_2014(self):
        self.model_circ_data_tt96("2014")
        py_b = get_data("circle_magpoles_5re_2014_km_bt96")
        idl_b = get_data("tst5re_2014_bt96")
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_tt01(self):
        tv_pos = get_data("tha_state_pos_gsm")
        t1 = tv_pos.times[0]
        t2 = tv_pos.times[-1]
        params = np.zeros([2, 10])
        params[:, 0] = self.pdyn
        params[:, 1] = self.dsti
        params[:, 2] = self.yimf
        params[:, 3] = self.zimf
        params[:, 4] = self.g1
        params[:, 5] = self.g2
        store_data("parmod", data={"x": [t1, t2], "y": params})
        tinterpol(
            "parmod", "tha_state_pos_gsm", method="nearest", newname="parmod_interp"
        )
        tt01("tha_state_pos_gsm", parmod="parmod_interp")
        py_b = get_data("tha_state_pos_gsm_bt01")
        idl_b = get_data("bt01")
        subtract("bt01", "tha_state_pos_gsm_bt01", "bt01_diff")
        tkm2re("tha_state_pos_gsm")
        tvectot("tha_state_pos_gsm_re", join_component=True)

        var_names = [
            "bt01",
            "tha_state_pos_gsm_bt01",
            "bt01_diff",
            "tha_state_pos_gsm_re_tot",
        ]
        local_png = os.path.join(save_dir, "t01_diffs.png")
        tplot(var_names, display=global_display, save_png=local_png)

        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_tts04(self):
        tv_pos = get_data("tha_state_pos_gsm")
        t1 = tv_pos.times[0]
        t2 = tv_pos.times[-1]
        params = np.zeros([2, 10])
        params[:, 0] = self.pdyn
        params[:, 1] = self.dsti
        params[:, 2] = self.yimf
        params[:, 3] = self.zimf
        params[:, 4] = self.w1
        params[:, 5] = self.w2
        params[:, 6] = self.w3
        params[:, 7] = self.w4
        params[:, 8] = self.w5
        params[:, 9] = self.w6
        store_data("parmod", data={"x": [t1, t2], "y": params})
        tinterpol(
            "parmod", "tha_state_pos_gsm", method="nearest", newname="parmod_interp"
        )
        tts04("tha_state_pos_gsm", parmod="parmod_interp")
        py_b = get_data("tha_state_pos_gsm_bts04")
        idl_b = get_data("bts04")
        subtract("bts04", "tha_state_pos_gsm_bts04", "bts04_diff")

        var_names = ["bts04", "tha_state_pos_gsm_bts04", "bts04_diff"]
        local_png = os.path.join(save_dir, "ts04_diffs.png")
        tplot(var_names, display=global_display, save_png=local_png)

        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_circ_tts04_2026(self):
        self.model_circ_data_tts04("2026")
        py_b = get_data("circle_magpoles_5re_2026_km_bts04")
        idl_b = get_data("tst5re_2026_bts04")
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_circ_tts04_2024(self):
        self.model_circ_data_tts04("2024")
        py_b = get_data("circle_magpoles_5re_2024_km_bts04")
        idl_b = get_data("tst5re_2024_bts04")
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_circ_tts04_2019(self):
        self.model_circ_data_tts04("2019")
        py_b = get_data("circle_magpoles_5re_2019_km_bts04")
        idl_b = get_data("tst5re_2019_bts04")
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_circ_tts04_2014(self):
        self.model_circ_data_tts04("2014")
        py_b = get_data("circle_magpoles_5re_2014_km_bts04")
        idl_b = get_data("tst5re_2014_bts04")
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_circ_tt01_2026(self):
        self.model_circ_data_tt01("2026")
        py_b = get_data("circle_magpoles_5re_2026_km_bt01")
        idl_b = get_data("tst5re_2026_bt01")
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_circ_tt01_2024(self):
        self.model_circ_data_tt01("2024")
        py_b = get_data("circle_magpoles_5re_2024_km_bt01")
        idl_b = get_data("tst5re_2024_bt01")
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_circ_tt01_2019(self):
        self.model_circ_data_tt01("2019")
        py_b = get_data("circle_magpoles_5re_2019_km_bt01")
        idl_b = get_data("tst5re_2019_bt01")
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)

    def test_circ_tt01_2014(self):
        self.model_circ_data_tt01("2014")
        py_b = get_data("circle_magpoles_5re_2014_km_bt01")
        idl_b = get_data("tst5re_2014_bt01")
        assert_allclose(py_b.y, idl_b.y, rtol=0.001, atol=0.5)


if __name__ == "__main__":
    unittest.main()



import logging
from .file_regex import maven_kp_l2_regex
from .config import CONFIG

from pyspedas.utilities.download import is_fsspec_uri
import fsspec

def get_filenames(query, public):
    """
    Retrieves file names based on the given query.

    Args:
        query (str): The query string used to filter the file names.
        public (bool): Flag indicating whether to use the public URL or private URL.

    Returns:
        str: The file names retrieved as a string.

    Raises:
        urllib.error.URLError: If there is an error in accessing the URL.
    """

    import urllib

    public_url = (
        "https://lasp.colorado.edu/maven/sdc/public/files/api/v1/search/science/fn_metadata/file_names"
        + "?"
        + query
    )
    private_url = (
        "https://lasp.colorado.edu/maven/sdc/service/files/api/v1/search/science/fn_metadata/file_names"
        + "?"
        + query
    )
    if not public:
        uname = CONFIG["maven_username"]
        pword = CONFIG["maven_password"]
        username = uname
        password = pword
        p = urllib.request.HTTPPasswordMgrWithDefaultRealm()
        p.add_password(None, private_url, username, password)
        handler = urllib.request.HTTPBasicAuthHandler(p)
        opener = urllib.request.build_opener(handler)
        urllib.request.install_opener(opener)
        logging.debug("get_filenames() making request to private URL: %s", private_url)
        page = urllib.request.urlopen(private_url)
        logging.debug("get_filenames() finished request to private URL: %s", private_url)

    else:
        logging.debug("get_filenames() making request to public URL: %s", public_url)
        page = urllib.request.urlopen(public_url)
        logging.debug("get_filenames() finished request to public URL: %s", public_url)

    return page.read().decode("utf-8")


def get_file_from_site(filename, public, data_dir):
    """
    Downloads a file from a website and saves it to the specified directory.

    Args:
        filename (str): The name of the file to download.
        public (bool): Indicates whether the file is public or private.
        data_dir (str): The directory where the file should be saved.

    Returns:
        None
    """
    import os
    import urllib

    public_url = (
        "https://lasp.colorado.edu/maven/sdc/public/files/api/v1/search/science/fn_metadata/download"
        + "?file="
        + filename
    )
    private_url = (
        "https://lasp.colorado.edu/maven/sdc/service/files/api/v1/search/science/fn_metadata/download"
        + "?file="
        + filename
    )

    if not public:
        uname = CONFIG["maven_username"]
        pword = CONFIG["maven_password"]
        username = uname
        password = pword
        p = urllib.request.HTTPPasswordMgrWithDefaultRealm()
        p.add_password(None, private_url, username, password)
        handler = urllib.request.HTTPBasicAuthHandler(p)
        opener = urllib.request.build_opener(handler)
        urllib.request.install_opener(opener)
        logging.debug("get_file_from_site making request to private url: %s", private_url)
        page = urllib.request.urlopen(private_url)
        logging.debug("get_file_from_site finished request to private url: %s", private_url)
    else:
        logging.debug("get_file_from_site making request to public url: %s", public_url)
        page = urllib.request.urlopen(public_url)
        logging.debug("get_file_from_site finished request to public url: %s", public_url)

    # Cloud Awareness: write page contents to URI
    if is_fsspec_uri(data_dir):
        protocol, path = data_dir.split("://")
        fs = fsspec.filesystem(protocol)

        fo = fs.open("/".join([data_dir, filename]), "wb")
    else:
        fo = open(os.path.join(data_dir, filename), "wb")

    with fo as code:
        code.write(page.read())

    return


def get_orbit_files():
    """
    Downloads MAVEN orbit files from the NASA NAIF website and saves them locally.

    Returns:
        None
    """
    import os
    import urllib
    import re

    orbit_files_url = "http://naif.jpl.nasa.gov/pub/naif/MAVEN/kernels/spk/"
    pattern = r">maven_orb_rec(\.orb|.{17}\.orb)<"
    logging.debug("get_orbit_files() making request to URL %s", orbit_files_url)
    page = urllib.request.urlopen(orbit_files_url)
    logging.debug("get_orbit_files() finished request to URL %s", orbit_files_url)
    page_string = str(page.read())
    toolkit_path = CONFIG["local_data_dir"]

    # Cloud Awareness: operate on fsspec filesystem instead of os
    if is_fsspec_uri(toolkit_path):
        protocol, path = toolkit_path.split("://")
        fs = fsspec.filesystem(protocol)

        orbit_files_path = "/".join([toolkit_path, "orbitfiles"])
        fs.mkdirs(orbit_files_path)
    else:
        orbit_files_path = os.path.join(toolkit_path, "orbitfiles")

        if not os.path.exists(toolkit_path):
            os.mkdir(toolkit_path)

        if not os.path.exists(orbit_files_path):
            os.mkdir(orbit_files_path)

    for matching_pattern in re.findall(pattern, page_string):
        filename = "maven_orb_rec" + matching_pattern
        logging.debug("get_orbit_files() making request to URL %s", orbit_files_url + filename)
        o_file = urllib.request.urlopen(orbit_files_url + filename)
        logging.debug("get_orbit_files() finished request to URL %s", orbit_files_url + filename)

        if is_fsspec_uri(toolkit_path):
            ifn = "/".join([orbit_files_path, filename])
            logging.debug("reading fsspec file %s",ifn)
            fo = fs.open(ifn, "wb")
        else:
            ifn = os.path.join(orbit_files_path, filename)
            logging.debug("reading plain file %s",ifn)
            fo = open(ifn, "wb")
        with fo as code:
            content=o_file.read()
            logging.debug("writing %d bytes into file %s",len(content), ifn)
            code.write(content)

    merge_orbit_files()

    return


def merge_orbit_files():
    """
    Merge MAVEN orbit files into a single file.

    This function searches for MAVEN orbit files in the 'orbitfiles' directory and merges them into a single file
    named 'merged_maven_orbits.orb' in the same directory. The files are sorted based on their dates before merging.

    Returns:
        None
    """
    import os
    import re

    toolkit_path = CONFIG["local_data_dir"]

    # Cloud Awareness: traverse fsspec-like filesystem instead of os
    if is_fsspec_uri(toolkit_path):
        protocol, path = toolkit_path.split("://")
        fs = fsspec.filesystem(protocol)

        orbit_files_path = "/".join([toolkit_path, "orbitfiles"])
        fl = fs.listdir(orbit_files_path, detail=False)
        fl = [f.rstrip("/").split("/")[-1] for f in fl]
        output_filename = "/".join([orbit_files_path, "merged_maven_orbits.orb"])
        fo = fs.open(output_filename, "w")
    else:
        orbit_files_path = os.path.join(toolkit_path, "orbitfiles")

        fl = os.listdir(orbit_files_path)
        output_filename = os.path.join(orbit_files_path,"merged_maven_orbits.orb")
        fo = open(output_filename, "w")

    pattern = "maven_orb_rec(_|)(|.{6})(|_.{9}).orb"
    orb_dates = []
    orb_files = []
    for f in fl:
        x = re.match(pattern, f)
        if x is not None:
            orb_file = os.path.join(orbit_files_path, f) if not is_fsspec_uri(toolkit_path) else "/".join([orbit_files_path, f])
            orb_files.append(orb_file)
            if x.group(2) != "":
                orb_dates.append(x.group(2))
            else:
                orb_dates.append("999999")

    sorted_files = [x for (y, x) in sorted(zip(orb_dates, orb_files))]

    with fo as code:
        skip_2_lines = False
        for o_file in sorted_files:
            logging.debug("merge_orbit_files processing file %s", o_file)
            if is_fsspec_uri(toolkit_path):
                # assumes fsspec filesystem triggered above
                fo_file = fs.open(o_file)
            else:
                fo_file = open(o_file)
            with fo_file as f:
                if skip_2_lines:
                    f.readline()
                    f.readline()
                skip_2_lines = True
                content=f.read()
                logging.debug("writing %d bytes to output file %s",len(content),output_filename)
                if type(content) is bytes:
                    code.write(str(content))
                else:
                    code.write(content)

    return


def get_access():
    """
    Check the access status from the access.txt file.

    Returns:
        bool: True if access is granted, False otherwise.
    """
    import os

    toolkit_path = os.path.dirname(__file__)
    with open(os.path.join(toolkit_path, "access.txt"), "r") as f:
        f.readline()
        s = f.readline().rstrip()
        s = s.split(" ")
        if s[1] == "1":
            return False
        else:
            return True


def get_root_data_dir():
    """
    Returns the preferred data download location for the pyspedas project.

    Raises:
        NameError: If "local_data_dir" is not found in config.py.

    Returns:
        str: The preferred data download location.
    """
    from .config import CONFIG

    # Get preferred data download location for pyspedas project
    if "local_data_dir" in CONFIG:
        return CONFIG["local_data_dir"]
    else:
        raise NameError("local_data_dir is not found in config.py")


def create_pref_file(toolkit_path, download_path):
    """
    Create a preferences file for the Maven toolkit.

    Args:
        toolkit_path (str): The path to the Maven toolkit.
        download_path (str): The path to the data download directory.

    Returns:
        None
    """
    import os

    # Put data download path into preferences file
    with open(os.path.join(toolkit_path, "mvn_toolkit_prefs.txt"), "w") as f:
        f.write("'; IDL Toolkit Data Preferences File'\n")
        f.write("mvn_root_data_dir: " + download_path)

    return


def set_new_data_root_dir():
    """
    Sets the new data root directory for the pyspedas project.

    This function prompts the user to enter a directory preference and validates
    if the specified path exists. If the path does not exist, the user is prompted
    to enter a new path until a valid path is provided. The function then sets the
    location of the mvn_toolkit_prefs file to the specified download path.

    Parameters:
        None

    Returns:
        None
    """
    import os

    # Get new preferred data download location for pyspedas project
    valid_path = input("Enter directory preference: ")

    # Cloud Awareness: had to rework for ensuring URI exists
    exists = False
    while not exists:
        if is_fsspec_uri(valid_path):
            if not "://" in valid_path:
                valid_path = input("Specified path does not exist. Enter new path: ")
            else:
                protocol, path = valid_path.split("://")
                fs = fsspec.filesystem(protocol)

                if fs.exists(valid_path):
                    exists = True
                else:
                    valid_path = input("Specified path does not exist. Enter new path: ")
        else:
            if os.path.exists(valid_path):
                exists = True
            else:
                valid_path = input("Specified path does not exist. Enter new path: ")
    download_path = valid_path
    logging.warning("Location of the mvn_toolkit_prefs file set to " + download_path)

    # Also edit the mvn_toolkit_prefs file to reflect the new data download location
    toolkit_path = CONFIG["local_data_dir"]
    create_pref_file(toolkit_path, download_path)


def get_new_files(files_on_site, data_dir, instrument, level):
    """
    Get the new files that are present on the site but not on the local hard drive.

    Parameters:
        files_on_site (list): List of files available on the site.
        data_dir (str): Path to the local data directory.
        instrument (str): Instrument name.
        level (str): Data level.

    Returns:
        list: List of new files that are present on the site but not on the local hard drive.
    """
    import os
    import re

    fos = files_on_site
    files_on_hd = []

    # Cloud Awareness: traverse fsspec filesystem
    if is_fsspec_uri(data_dir):
        protocol, path = data_dir.split("://")
        fs = fsspec.filesystem(protocol)

        walk = fs.walk(data_dir)
    else:
        walk = os.walk(data_dir)

    for dir, _, files in walk:
        for f in files:
            if re.match("mvn_" + instrument + "_" + level + "_*", f):
                files_on_hd.append(f)

    x = set(files_on_hd).intersection(files_on_site)
    for matched_file in x:
        fos.remove(matched_file)

    return fos


def create_dir_if_needed(f, data_dir, level):
    """
    Create a directory if it does not already exist based on the year and month extracted from the file name.

    Parameters:
        f (str): The file name.
        data_dir (str): The base directory where the new directory will be created.
        level (str): The level of the file ("insitu" or "sci").

    Returns:
        str: The full path of the created directory.
    """
    import os

    if level == "insitu":
        year, month, _ = get_year_month_day_from_kp_file(f)
    else:
        year, month, _ = get_year_month_day_from_sci_file(f)

    # Cloud Awareness: operate upon fsspec-like filesystem
    if is_fsspec_uri(data_dir):
        protocol, path = data_dir.split("://")
        fs = fsspec.filesystem(protocol)

        full_path = "/".join([data_dir, year, month])
        fs.mkdirs(full_path)
        return full_path

    if not os.path.exists(os.path.join(data_dir, year, month)):
        os.makedirs(os.path.join(data_dir, year, month))

    full_path = os.path.join(data_dir, year, month)

    return full_path


def get_year_month_day_from_kp_file(f):
    """
    Extracts the year, month, and day from a given file name.

    Args:
        f (str): The file name.

    Returns:
        tuple: A tuple containing the year, month, and day extracted from the file name.
    """
    date_string = f.split("_")[3]
    year = date_string[0:4]
    month = date_string[4:6]
    day = date_string[6:8]

    return year, month, day


def get_year_month_day_from_sci_file(f):
    """
    Extracts the year, month, and day from a given scientific file name.

    Parameters:
        f (str): The scientific file name.

    Returns:
        tuple: A tuple containing the year, month, and day extracted from the file name.
    """
    kp_regex, l2_regex = maven_kp_l2_regex()
    m = l2_regex.match(f)
    if m is None:
        logging.error('l2_regex match failed for filename %s', f)
    year = m.group("year")
    month = m.group("month")
    day = m.group("day")

    return year, month, day


def display_progress(x, y):
    """
    Display the progress of a task as a visual bar.

    Parameters:
    x (int): The current progress value.
    y (int): The total value representing the completion of the task.

    Returns:
    None
    """
    num_stars = int(round(float(x) / y * 70))
    logging.warning(
        "||"
        + "*" * num_stars
        + "-" * (70 - num_stars)
        + "||"
        + " ( "
        + str(round(100 * float(x) / y))
        + "% )"
    )
    return


def get_uname_and_password():
    """
    Prompts the user to enter their username and password to access the team website.

    Returns:
        None
    """
    import getpass

    uname = input("Enter user name to access the team website: ")
    pword = getpass.getpass("Enter your password: ")

    CONFIG["maven_username"] = uname
    CONFIG["maven_password"] = pword
    return



import logging
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG


def load(
    trange=["1998-04-06", "1998-04-07"],
    instrument="mam",
    datatype="pp",
    prefix="",
    suffix="",
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=False,
    force_download=False,
):
    """
    Load data from the Equator-S mission.

    Parameters
    ----------
    trange : list of str, optional
        Time range of interest [starttime, endtime] with the format
        ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
        ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss'].
        Default is ['1998-04-06', '1998-04-07'].
    instrument : str, list of str, optional
        Type of instrument.
        Valid values: 'aux', 'edi', 'epi',  'ici', 'mam', 'pcd', 'sfd', 'all'.
        Default is 'mam'.
        If 'all' is specified, all instruments will be loaded.
    datatype : str, optional
        Type of data.
        Default is 'pp'.
        Valid values:
        -- 'pp' for instruments: 'aux', 'edi', 'epi',  'ici', 'mam', 'pcd'
        -- 'sp' for instruments: 'sfd', 'edi'
        Instrument 'edi' has both 'pp' and 'sp' data; for all other instruments, datatype is ignored.
    prefix: str, optional
        The tplot variable names will be given this prefix.
        Default is ''.
    suffix : str, optional
        The tplot variable names will be given this suffix.
        Default is ''.
    get_support_data : bool, optional
        Data with an attribute "VAR_TYPE" with a value of "support_data" will be loaded into tplot.
        Default is False, only loads in data with a "VAR_TYPE" attribute of "data".
    varformat : str, optional
        The file variable formats to load into tplot.
        Wildcard character "*" is accepted.
        Default is None, all variables are loaded in.
    varnames : list of str, optional
        List of variable names to load.
        Default is [], all data variables are loaded.
    downloadonly : bool, optional
        Set this flag to download the CDF files, but not load them into tplot variables.
        Default is False.
    notplot : bool, optional
        Return the data in hash tables instead of creating tplot variables.
        Default is False.
    no_update : bool, optional
        If set, only load data from your local cache.
        Default is False.
    time_clip : bool, optional
        Time clip the variables to exactly the range specified in the trange keyword.
        Default is False.
    force_download : bool, optional
        Download file even if local version is more recent than server version.
        Default is False.

    Returns
    -------
    list of str/dictionary
        List of tplot variables created.
        If downloadonly is set to True, returns a list of the downloaded files.
        If notplot is set to True, returns a dictionary of the data loaded.


    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> edi_vars = pyspedas.projects.equator_s.edi(trange=['1998-04-06', '1998-04-07'])
    >>> print(edi_vars)
    ['V_ed_xyz_gse%eq_pp_edi', 'E_xyz_gse%eq_pp_edi']
    >>> tplot(edi_vars)
    """

    out_files = []
    out_vars = []

    if (
        trange is None
        or not isinstance(trange, list)
        or len(trange) != 2
        or trange[0] > trange[1]
    ):
        logging.error("Invalid trange specified.")
        return out_vars

    if not isinstance(instrument, list):
        instrument = [instrument]
    if "all" in instrument:
        instrument = ["aux", "edi", "epi", "ici", "mam", "pcd", "sfd"]

    for instr in instrument:

        if instr == "aux":
            pathformat = "pp/aux/%Y/eq_pp_aux_%Y%m%d_v??.cdf"
        elif instr == "edi":
            # this is the only case that datatype is not predetermined
            # it can be either 'pp' or 'sp'
            if datatype == "sp":
                pathformat = "sp/edi/%Y/eq_sp_edi_%Y%m%d_v??.cdf"
            else:
                pathformat = "pp/edi/%Y/eq_pp_edi_%Y%m%d_v??.cdf"
        elif instr == "epi":
            pathformat = "pp/epi/%Y/eq_pp_epi_%Y%m%d_v??.cdf"
        elif instr == "ici":
            pathformat = "pp/ici/%Y/eq_pp_ici_%Y%m%d_v??.cdf"
        elif instr == "mam":
            pathformat = "pp/mam/%Y/eq_pp_mam_%Y%m%d_v??.cdf"
        elif instr == "pcd":
            pathformat = "pp/pcd/%Y/eq_pp_pcd_%Y%m%d_v??.cdf"
        elif instr == "sfd":
            # Only datatype='sp' is available for SFD
            pathformat = "sp/sfd/%Y/eq_sp_sfd_%Y%m%d_v??.cdf"
        else:
            logging.error("Invalid instrument type." + instr)
            continue

        # find the full remote path names using the trange
        remote_names = dailynames(file_format=pathformat, trange=trange)

        # download the files
        files = download(
            remote_file=remote_names,
            remote_path=CONFIG["remote_data_dir"],
            local_path=CONFIG["local_data_dir"],
            no_download=no_update,
            force_download=force_download,
        )
        if files is not None:
            if not isinstance(files, list):
                files = [files]
            out_files.extend(files)

        if not downloadonly and len(files) > 0:
            # Read the files into tplot variables
            vars = cdf_to_tplot(
                files,
                prefix=prefix,
                suffix=suffix,
                get_support_data=get_support_data,
                varformat=varformat,
                varnames=varnames,
                notplot=notplot,
            )
            if not isinstance(vars, list):
                vars = [vars]
            out_vars.extend(vars)

    out_files = list(set(out_files))
    out_files = sorted(out_files)

    if not downloadonly and len(out_files) < 1:
        logging.info("No files were downloaded.")
        return out_files  # return an empty list

    if downloadonly:
        return out_files

    if notplot:
        if len(out_vars) < 1:
            logging.info("No variables were loaded.")
            return {}
        else:
            return out_vars[0]  # return data in hash tables

    if time_clip:
        tclip(out_vars, trange[0], trange[1], suffix="", overwrite=True)

    return out_vars



import logging
from pyspedas.tplot_tools import get_data, store_data
import xarray as xr
from packaging.version import Version

def tdotp(variable1, variable2, newname=None):
    """
    Calculate the sample-by-sample dot product of two tplot variables

    Parameters
    -----------
    variable1: str
        First tplot variable
    variable2: str
        Second tplot variable
    newname: str
        Name of the output variable
        Default: variable1_dot_variable2
            
    Returns
    --------
    str
        Name of the tplot variable created

    Examples
    --------

        >>> # Compute the dot products of a given time series
        >>> import pyspedas
        >>> x1 = [0, 4, 8]
        >>> x2 = [0, 4, 8]
        >>> time1 = [pyspedas.time_float("2020-01-01") + i for i in x1]
        >>> time2 = [pyspedas.time_float("2020-01-01") + i for i in x2]
        >>> pyspedas.store_data("a", data={"x": time1, "y": [[1, 2, 3],[2, 3, 4],[3, 4, 5]]})
        >>> pyspedas.store_data("c", data={"x": time2, "y": [[1, 4, 1],[2, 5, 2],[3, 5, 3]]})
        >>> n = pyspedas.tdotp("a", "c", newname="a_dot_c")
        >>> print('new tplot variable:', n)
        >>> ac = pyspedas.get_data(n)
        >>> print(ac)

    """

    data1 = get_data(variable1, xarray=True)
    data2 = get_data(variable2, xarray=True)

    if data1 is None:
        logging.error('Variable not found: ' + variable1)
        return

    if data2 is None:
        logging.error('Variable not found: ' + variable2)
        return

    if newname is None:
        newname = variable1 + '_dot_' + variable2

    # Check xarray version to see whether it wants 'dims' (older releases) or 'dim' (2023.12.0 and later)
    # Note that with semantic versioning, we can't just do string comparisons, so we use packaging.version.Version

    # calculate the dot product
    if Version(xr.__version__) < Version('2023.12.0'):
        # old way
        out = data1.dot(data2, dims='v_dim')
    else:
        # new way
        out = data1.dot(data2, dim='v_dim')


    # save the output
    saved = store_data(newname, data={'x': data1.time.values, 'y': out.values})

    if saved is not None:
        return newname



import datetime
import logging
import numpy as np
from pyspedas.tplot_tools import time_double
from pyspedas.utilities.interpol import interpol


def time_ephemeris(t, et2ut=False):
    """
    Convert between unix time and ephemeris time, based on the IDL routine by Davin Larson

    Parameters
    ==========

    t: float
        Input time to convert, by default unix time (seconds since 1970-01-01)
    et2ut: bool
        If True, convert ephemeris time (seconds since 2000-01-01/11:58:55 UTC) to unix time (Default: False)

    Examples
    ========

    >>> strdate='2025-01-01/00:00:00'
    >>> ut = pyspedas.time_double(strdate)
    >>> et = pyspedas.time_ephemeris(ut)
    >>> print(et)
    788961669.184
    """
    if not isinstance(t, float):
        t = time_double(t)

    ls_utimes = time_double(
        ['0200-1-1', '1972-1-1', '1972-7-1', '1973-1-1', '1974-1-1', '1975-1-1', '1976-1-1', '1977-1-1', '1978-1-1',
         '1979-1-1', '1980-1-1', '1981-7-1', '1982-7-1', '1983-7-1', '1985-7-1', '1988-1-1', '1990-1-1', '1991-1-1', '1992-7-1', '1993-7-1', '1994-7-1',
         '1996-1-1', '1997-7-1', '1999-1-1', '2006-1-1', '2009-1-1', '2012-7-1', '2015-7-1', '2017-1-1', '3000-1-1'])

    ls_num = np.arange(len(ls_utimes)) + 9
    utc_et_diff = time_double('2000-1-1/12:00:00') - 32.184
    ls_etimes = ls_utimes + ls_num - utc_et_diff
    disable_time = time_double('2027-07-01')  # time of next possible leap second

    if time_double() > disable_time - 30*86400.0:
        logging.warning('Warning: This procedure must be modified before ' + str(disable_time) + ' to account for potential leap second')

    if time_double() > disable_time:
        raise ValueError('Sorry!  This procedure has been disabled because it was not modified to account for a possible leap second on ' + str(disable_time))

    if time_double() > disable_time - 7*86400.0:
        logging.warning('Warning: This procedure must be modified before ' + str(disable_time) + ' to account for potential leap second at that time.')

    if et2ut:
        return t - np.floor(interpol(ls_num, ls_etimes, t)) + utc_et_diff

    ut = time_double(t)

    return ut + np.floor(interpol(ls_num, ls_utimes, ut)) - utc_et_diff




import logging
from time import time
from pyspedas.projects import mms
from pyspedas.projects.mms.particles.mms_part_products import mms_part_products
from pyspedas.tplot_tools import time_double
logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def mms_part_getspec(instrument='fpi',
                     probe='1',
                     species='e',
                     data_rate='fast',
                     trange=None,
                     output=['energy', 'theta', 'phi', 'pa', 'gyro'],
                     units='eflux',
                     energy=None,
                     phi=None,
                     theta=None,
                     pitch=None,
                     gyro=None,
                     mag_name=None,
                     mag_data_rate=None,
                     pos_name=None,
                     sc_pot_name=None,
                     sc_pot_data_rate=None,
                     fac_type='mphigeo',
                     center_measurement=False,
                     spdf=False,
                     correct_photoelectrons=False,
                     internal_photoelectron_corrections=False,
                     disable_photoelectron_corrections=False,
                     zero_negative_values=False,
                     regrid=[32, 16],
                     no_regrid=False,
                     prefix='',
                     suffix='', 
                     subtract_bulk = False,
                     vel_name = None,
                     vel_data_rate = None,
                     sdc_units = False,
                     ):
    """
    Generate spectra and moments from 3D MMS particle data

    Parameters
    ----------
        trange: list of str
            Time range

        units: str
            Specify units of output variables; must be 'eflux' to calculate moments.   Valid options::

                'flux'   -  # / (cm^2 * s * sr * eV)
                'eflux'  -  eV / (cm^2 * s * sr * eV)  <default>
                'df_cm'  -  s^3 / cm^6
                'df_km'  -  s^3 / km^6

            Default: 'eflux'

        species: str
            Specify the species of the input tplot variable (Default: 'e')

        data_rate: str
            Data rate of the input data (Default: 'fast')

        instrument: str
            Instrument (FPI or HPCA)

        probe: int or str
            Spacecraft probe # (Default: 1)

        output: str or list of str
            Output variables; options::

                'energy': energy spectrograms
                'theta': theta spectrograms
                'phi': phi spectrograms
                'pa': pitch-angle spectrograms
                'gyro': gyro-phase spectrograms
                'moments': plasma moments

            Default: ['energy', 'theta', 'phi', 'pa', 'gyro']

        energy: list of float
            Energy range [min, max], in eV

        phi: list of float
            Phi range [min, max], in degrees

        theta: list of float
            Theta range [min, max], in degrees

        pitch: list of float
            Pitch-angle range [min, max], in degrees

        gyro: list of float
            Gyro-angle range [min, max], in degrees

        mag_name: str
            Tplot variable containing magnetic field data for
            moments and FAC transformations

        mag_data_rate: str
            Data rate to use when loading the spacecraft potential data ('fast', 'srvy' or 'brst')
            If not given, defaults to value of 'data_rate'.

        pos_name: str
            Tplot variable containing spacecraft position for
            FAC transformations. If not given, defaults to 'mmsN_mec_r_gse'

        sc_pot_name: str
            If set, overrides default sc_pot variable name constructed from sc_pot_data_rate below.
            Default: None

        sc_pot_data_rate: str
            Data rate to use when loading the spacecraft potential data ('fast', 'srvy' or 'brst')
            If not given, defaults to value of 'data_rate'.

        fac_type: str
            Field aligned coordinate system variant; options::

                'phigeo'
                'mphigeo'
                'xgse'

            Default: 'mphigeo'

        ### ADDED 4.16.2025, TESTING BULK VELOCITY SUBTRACTION
        subtract_bulk: bool
            subtract the bulk velocity prior to doing the calculations

        vel_name: str
            Tplot variable containing velocity data in km/s for use with /subtract_bulk

        vel_data_rate: str
            Data rate to use when loading the bulk velocity data ('fast' or 'brst')

        #################
            
        correct_photoelectrons: bool
            Flag to correct FPI data for photoelectrons 
            (defaults to True for FPI electron data - disable with the parameter below)

        disable_photoelectron_corrections: bool
            Flag to disable FPI photoelectron corrections

        internal_photoelectron_corrections: bool
            Apply internal photoelectron corrections

        zero_negative_values: bool
            Turn negative values to 0 after doing the photoelectron corrections (DES)

        regrid: array of int
            Dimensions for regridded data array (see no_regrid flag below)
            Default: [32,16]

        no_regrid: bool
            Flag to disable regridding of data array

        prefix: str
            Prefix for output tplot variables. Default: ''

        suffix: str
            Suffix for output tplot variables. Default: ''

        sdc_units: bool
            If True, convert moments ptens and qflux values and units to nPa and mW/m^2 respectively, for
            compatibility with MMS SDC products
            Default: False

    Returns
    -------
    None
       Creates tplot variables containing spectrograms and moments

    """

    start_time = time()

    if trange is None:
        logging.error('Time range not specified; please specify time range using the trange keyword.')
        return

    if vel_data_rate is None:  # type: ignore
        if data_rate == 'brst':
            vel_data_rate = 'brst'
        else:
            vel_data_rate = 'srvy'

    if mag_data_rate is None:
        if data_rate == 'brst':
            mag_data_rate = 'brst'
        else:
            mag_data_rate = 'srvy'

    if sc_pot_data_rate is None:
        if data_rate == 'brst':
            sc_pot_data_rate = 'brst'
        else:
            sc_pot_data_rate = 'fast'

    instrument = instrument.lower()

    # HPCA is required to be at the center of the accumulation interval
    # due to assumptions made in mms_get_hpca_dist
    if instrument == 'hpca' and center_measurement == False:
        center_measurement = True
    
    if instrument == 'fpi':
        data_vars = mms.fpi(datatype='d'+species+'s-dist', probe=probe, data_rate=data_rate, trange=trange, time_clip=True, center_measurement=center_measurement, spdf=spdf)

        # load bulk velocity variables if user requests to subtract
        if subtract_bulk == True and vel_name is None: 
            vel_vars = mms.fpi(datatype='d'+species+'s-moms', probe=probe, data_rate=data_rate, trange=trange, varformat='*_d'+species+'s_bulkv_gse_*' + '|' + '*_d'+species+'*s_bulkv_spintone_gse_*', spdf=spdf)
            if not vel_vars:
                logging.error('Error, no bulk velocity data loaded.')
                return
        
    elif instrument == 'hpca':
        # for HPCA, 'fast' should be 'srvy'
        if data_rate == 'fast':
            data_rate = 'srvy'
        # 'i' and 'e' are only valid for FPI
        if species in ['i', 'e']:
            species = 'hplus'
        data_vars = mms.hpca(datatype='ion', probe=probe, data_rate=data_rate, trange=trange, time_clip=True, center_measurement=center_measurement, get_support_data=True, spdf=spdf)

        # load bulk velocity variables if user requests to subtract
        if subtract_bulk == True and vel_name is None:
            vel_vars = mms.hpca(datatype='moments', probe=probe, data_rate=data_rate, trange=trange, varformat='*_hpca_'+species+'_ion_bulk_velocity', spdf=spdf)
            if not vel_vars:
                logging.error('Error, no bulk velocity data loaded.')
                return
    else:
        logging.error('Error, unknown instrument: ' + instrument + '; valid options: fpi, hpca')
        return

    if data_vars is None or len(data_vars) == 0:
        logging.error('Error, no data loaded.')
        return None

    if not isinstance(probe, list):
        probe = [probe]

    if instrument == 'fpi' and species == 'e' and 'moments' in output and not disable_photoelectron_corrections:
        correct_photoelectrons = True

    support_trange = [time_double(trange[0])-60.0, time_double(trange[1])+60.0]

    
    # load state data (needed for coordinate transformations and field-aligned coordinates)
    pos_vars = mms.mec(probe=probe, trange=support_trange, time_clip=True, spdf=spdf)

    if len(pos_vars) == 0:
        logging.error('Error, no state data loaded.')
        return

    mag_vars = mms.fgm(probe=probe, trange=support_trange, data_rate=mag_data_rate, time_clip=True, spdf=spdf)

    if len(mag_vars) == 0:
        logging.error('Error, no magnetic field data loaded.')
        return

    scpot_vars = mms.edp(probe=probe, trange=support_trange, level='l2', spdf=spdf, data_rate=sc_pot_data_rate, datatype='scpot', varformat='*_edp_scpot_*')

    out_vars = []

    for prb in probe:
        prb_str = str(prb)

        #### new #####

        if vel_name is None or vel_name == '':
            if instrument == 'fpi':
                vel_name = 'mms'+prb_str+'_d'+species+'s_bulkv_gse_'+vel_data_rate
            else:
                vel_name = 'mms'+prb_str+'_hpca_'+species+'_ion_bulk_velocity'
        else: 
            logging.info("Using non-default variable %s for vel data",vel_name)

        ###### new #######

        if mag_name is None or mag_name == '':
            mag_name = 'mms'+prb_str+'_fgm_b_gse_'+mag_data_rate+'_l2_bvec'
        else:
            logging.info("Using non-default variable %s for mag data",mag_name)

        if pos_name is None or pos_name == '':
            pos_name = 'mms'+prb_str+'_mec_r_gse'
        else:
            logging.info("Using non-default variable %s for position data",pos_name)

        if sc_pot_name is None or sc_pot_name == '':
            sc_pot_name = 'mms' + prb_str + '_edp_scpot_' + sc_pot_data_rate + '_l2'
        else:
            logging.info('Using non-default variable %s for sc_pot data', sc_pot_name)

        if instrument == 'fpi':
            tname = 'mms'+prb_str+'_d'+species+'s_dist_'+data_rate
        elif instrument == 'hpca':
            tname = 'mms'+prb_str+'_hpca_'+species+'_phase_space_density'

        new_vars = mms_part_products(tname, species=species, instrument=instrument, probe=prb, data_rate=data_rate,
                          output=output, units=units, energy=energy, phi=phi, theta=theta, pitch=pitch, gyro=gyro,
                          mag_name=mag_name, pos_name=pos_name, vel_name = vel_name, fac_type=fac_type, sc_pot_name=sc_pot_name,
                          correct_photoelectrons=correct_photoelectrons, zero_negative_values=zero_negative_values,
                          internal_photoelectron_corrections=internal_photoelectron_corrections,
                          disable_photoelectron_corrections=disable_photoelectron_corrections, regrid=regrid,
                          no_regrid=no_regrid, prefix=prefix, suffix=suffix, subtract_bulk = subtract_bulk,
                          sdc_units=sdc_units)
        
        if new_vars is None:
            continue

        out_vars = out_vars + new_vars

    logging.info('Finished; time to run: ' + str(round(time()-start_time, 1)) + ' seconds.')
    
    return out_vars



from functools import partial
from typing import TypeVar

# Credit to Aidan Grant, from the comments of a Jetbrains open issue (linked below) regarding autocompletion
# of functions defined with functools.partial.  This seems to solve the issue with
# PyCharm not being able to autocomplete or show arguments for some of our load routines,
# e.g. from pyspedas.projects.goes

T = TypeVar('T')

def better_partial(func: T, *args, **kwargs) -> T:
    """
    Behaves the same as functools.partial, but works with idea's type hinting.
    @see https://youtrack.jetbrains.com/issue/PY-25698/No-proper-completion-or-parameter-info-for-partially-applied-parameters-of-functools.partial
    """
    return partial(func, *args, **kwargs)



"""Tests of twavpol functions."""

import os
import logging
import unittest
import numpy as np
from numpy.testing import assert_allclose
import pyspedas
from pyspedas import get_data, tplot, tplot_rename, tplot_copy, twavpol
from pyspedas.utilities.config_testing import TESTING_CONFIG, test_data_download_file

# Whether to display plots during testing
global_display = TESTING_CONFIG["global_display"]
# Directory to save testing output files
output_dir = TESTING_CONFIG["local_testing_dir"]
# Ensure output directory exists
analysis_dir = "analysis_tools"
save_dir = os.path.join(output_dir, analysis_dir)
if not os.path.exists(save_dir):
    os.makedirs(save_dir)
# Directory with IDL SPEDAS validation files
validation_dir = TESTING_CONFIG["remote_validation_dir"]


class TwavpolDataValidation(unittest.TestCase):
    """Compares cotrans results between Python and IDL"""

    @classmethod
    def setUpClass(cls):
        """
        IDL Data has to be downloaded to perform these tests
        The SPEDAS script that creates the file: general/tools/python_validate/python_wavpol_validate.pro
        """
        # Testing tolerance
        cls.tol = 1e-10

        # Load IDL savefile
        filename = test_data_download_file(
            validation_dir, analysis_dir, "thc_twavpol_validate.tplot", save_dir
        )
        if not filename:
            # Skip tests
            raise unittest.SkipTest("Cannot download data validation file")

        # Load validation variables from the test file
        pyspedas.del_data("*")
        # pyspedas.cdf_to_tplot(filename)
        pyspedas.tplot_restore(filename)
        pyspedas.tplot_names()
        # pyspedas.tplot('thc_scf_fac')
        # pyspedas.tplot('thc_scf_fac_powspec')
        # pyspedas.tplot('thc_scf_fac')
        cls.thc_scf_fac = pyspedas.get_data("thc_scf_fac")
        cls.thc_scf_fac_attr = pyspedas.get_data("thc_scf_fac", metadata=True)
        cls.thc_scf_fac_powspec = pyspedas.get_data("thc_scf_fac_powspec")
        cls.thc_scf_fac_powspec_attr = pyspedas.get_data(
            "thc_scf_fac_powspec", metadata=True
        )
        cls.thc_scf_fac_degpol = pyspedas.get_data("thc_scf_fac_degpol")
        cls.thc_scf_fac_waveangle = pyspedas.get_data("thc_scf_fac_waveangle")
        cls.thc_scf_fac_elliptict = pyspedas.get_data("thc_scf_fac_elliptict")
        cls.thc_scf_fac_helict = pyspedas.get_data("thc_scf_fac_helict")
        cls.thc_scf_fac_pspec3 = pyspedas.get_data("thc_scf_fac_pspec3")

        tplot_copy("thc_scf_fac", "idl_thc_scf_fac")
        tplot_rename("thc_scf_fac_powspec", "idl_thc_scf_fac_powspec")
        tplot_rename("thc_scf_fac_degpol", "idl_thc_scf_fac_degpol")
        tplot_rename("thc_scf_fac_waveangle", "idl_thc_scf_fac_waveangle")
        tplot_rename("thc_scf_fac_elliptict", "idl_thc_scf_fac_elliptict")
        tplot_rename("thc_scf_fac_helict", "idl_thc_scf_fac_helict")
        tplot_rename("thc_scf_fac_pspec3", "idl_thc_scf_fac_pspec3")

        twavpol("thc_scf_fac")

    def setUp(self):
        """We need to clean tplot variables before each run"""
        # pyspedas.del_data('*')

    def test_multiple_twavpol_call(self):
        """Validate twavpol power spectrum output between two calls with the same input"""

        twavpol("thc_scf_fac")
        before_py_powspec = get_data("thc_scf_fac_powspec")
        twavpol("thc_scf_fac")
        after_py_powspec = get_data("thc_scf_fac_powspec")
        assert_allclose(before_py_powspec.y, after_py_powspec.y)
        logging.info(
            "nanmin/nanmax of powspec: %f %f ",
            np.nanmin(after_py_powspec.y),
            np.nanmax(after_py_powspec.y),
        )
        assert_allclose(
            after_py_powspec.times, self.thc_scf_fac_powspec.times, atol=1.0e-06
        )
        assert_allclose(
            after_py_powspec.y, self.thc_scf_fac_powspec.y, atol=1.0e-06, rtol=1.0e-06
        )
        local_png = os.path.join(save_dir, "multiple_twavpol.png")
        tplot(
            ["thc_scf_fac_powspec", "idl_thc_scf_fac_powspec"],
            display=global_display,
            save_png=local_png,
        )

    def test_powspec(self):
        """Validate twavpol power spectrum output"""

        py_powspec = get_data("thc_scf_fac_powspec")
        logging.info(
            "nanmin/nanmax of powspec: %f %f ",
            np.nanmin(py_powspec.y),
            np.nanmax(py_powspec.y),
        )
        assert_allclose(py_powspec.times, self.thc_scf_fac_powspec.times, atol=1.0e-06)
        assert_allclose(
            py_powspec.y, self.thc_scf_fac_powspec.y, atol=1.0e-06, rtol=1.0e-06
        )
        local_png = os.path.join(save_dir, "powspec.png")
        tplot(
            ["thc_scf_fac_powspec", "idl_thc_scf_fac_powspec"],
            display=global_display,
            save_png=local_png,
        )

    def test_degpol(self):
        """Validate twavpol degpol output"""

        py_degpol = pyspedas.get_data("thc_scf_fac_degpol")
        logging.info(
            "nanmin/nanmax of degpol: %f %f ",
            np.nanmin(py_degpol.y),
            np.nanmax(py_degpol.y),
        )
        assert_allclose(py_degpol.times, self.thc_scf_fac_degpol.times, atol=1.0e-06)
        assert_allclose(
            py_degpol.y, self.thc_scf_fac_degpol.y, atol=1.0e-06, rtol=1.0e-06
        )
        local_png = os.path.join(save_dir, "degpol.png")
        tplot(
            ["thc_scf_fac_degpol", "idl_thc_scf_fac_degpol"],
            display=global_display,
            save_png=local_png,
        )

    def test_waveangle(self):
        """Validate twavpol waveangle output"""

        py_waveangle = pyspedas.get_data("thc_scf_fac_waveangle")
        logging.info(
            "nanmin/nanmax of waveangle: %f %f ",
            np.nanmin(py_waveangle.y),
            np.nanmax(py_waveangle.y),
        )
        assert_allclose(
            py_waveangle.times, self.thc_scf_fac_waveangle.times, atol=1.0e-05
        )
        assert_allclose(
            py_waveangle.y, self.thc_scf_fac_waveangle.y, atol=1.0e-05, rtol=1.0e-06
        )
        local_png = os.path.join(save_dir, "waveangle.png")
        tplot(
            ["thc_scf_fac_waveangle", "idl_thc_scf_fac_waveangle"],
            display=global_display,
            save_png=local_png,
        )

    def test_elliptict(self):
        """Validate twavpol elliptict output"""

        py_elliptict = pyspedas.get_data("thc_scf_fac_elliptict")
        logging.info(
            "nanmin/nanmax of elliptict: %f %f ",
            np.nanmin(py_elliptict.y),
            np.nanmax(py_elliptict.y),
        )
        assert_allclose(
            py_elliptict.times, self.thc_scf_fac_elliptict.times, atol=1.0e-06
        )
        assert_allclose(
            py_elliptict.y, self.thc_scf_fac_elliptict.y, atol=1.0e-06, rtol=1.0e-06
        )
        local_png = os.path.join(save_dir, "elliptict.png")
        tplot(
            ["thc_scf_fac_elliptict", "idl_thc_scf_fac_elliptict"],
            display=global_display,
            save_png=local_png,
        )

    def test_helict(self):
        """Validate twavpol helict output"""

        py_helict = pyspedas.get_data("thc_scf_fac_helict")
        logging.info(
            "nanmin/nanmax of helict: %f %f ",
            np.nanmin(py_helict.y),
            np.nanmax(py_helict.y),
        )
        assert_allclose(py_helict.times, self.thc_scf_fac_helict.times, atol=1.0e-06)
        assert_allclose(
            py_helict.y, self.thc_scf_fac_helict.y, atol=1.0e-06, rtol=1.0e-06
        )
        local_png = os.path.join(save_dir, "helict.png")
        tplot(
            ["thc_scf_fac_helict", "idl_thc_scf_fac_helict"],
            display=global_display,
            save_png=local_png,
        )

    def test_pspec3(self):
        """Validate twavpol pspec3 output"""

        py_pspec3 = pyspedas.get_data("thc_scf_fac_pspec3")
        logging.info(
            "nanmin/nanmax of pspec3: %f %f ",
            np.nanmin(py_pspec3.y),
            np.nanmax(py_pspec3.y),
        )
        assert_allclose(py_pspec3.times, self.thc_scf_fac_pspec3.times, atol=1.0e-06)
        assert_allclose(
            py_pspec3.y, self.thc_scf_fac_pspec3.y, atol=1.0e-06, rtol=1.0e-06
        )
        local_png = os.path.join(save_dir, "pspec3.png")
        tplot(
            ["thc_scf_fac_pspec3", "idl_thc_scf_fac_pspec3", "thc_scf_fac_pspec3*"],
            display=global_display,
            save_png=local_png,
        )

    @unittest.skip("skipping, work in progress")
    def test_mms_scm(self):
        # achDate = ['2015-09-19/10:07:00', '2015-09-19/10:07:12']
        achDate = ["2015-09-19/10:06:00", "2015-09-19/10:09:00"]
        SCMbrst_vars = pyspedas.projects.mms.scm(
            probe=4, data_rate="brst", trange=achDate, time_clip=True
        )
        logging.info("SCM burst variables: %s", SCMbrst_vars)
        # SCM burst sampling rate = 8192/inputs as done in example tutorial
        nopfft_input = 8192  # number of points for FFT
        steplength_input = (
            nopfft_input / 2
        )  # number of points for shifting between 2 FFT
        bin_freq_input = 3  # number of bins for frequency averaging
        pyspedas.twavpol(
            "mms4_scm_acb_gse_scb_brst_l2",
            nopfft=nopfft_input,
            steplength=steplength_input,
            bin_freq=bin_freq_input,
        )

        local_png = os.path.join(save_dir, "mms_scm_burst.png")
        tplot(
            [
                "mms4*scb_brst_l2",
                "mms4*powspec*",
                "mms4*degpol*",
                "mms4*elliptc",
                "mms4*helict",
                "mms4*angle",
            ],
            display=global_display,
            save_png=local_png,
        )


if __name__ == "__main__":
    unittest.main()



import numpy as np
import unittest
from pyspedas.projects.mms.particles.mms_part_slice2d import mms_part_slice2d
from pyspedas.particles.spd_units_string import spd_units_string
global_display = False

class SliceTests(unittest.TestCase):
    def test_notime(self):
        nothing = mms_part_slice2d()
        self.assertTrue(not nothing)

    def test_fpi_nospecies(self):
        mms_part_slice2d(time='2015-10-16/13:06:30', probe='1', rotation='xy', display=global_display)
        mms_part_slice2d(time='2015-10-16/13:06:30', probe='1', instrument='hpca', rotation='xy', display=global_display)

    def test_fpi_return_slice(self):
        the_slice = mms_part_slice2d(time='2015-10-16/13:06:30', probe='1', rotation='xy', display=global_display, return_slice=True)
        self.assertTrue(isinstance(the_slice, dict))

    def test_fpi_brst_slice_x(self):
        mms_part_slice2d(time='2015-10-16/13:06:30', slice_x=np.array([1, 0, 0]), probe='1', species='i', data_rate='brst', rotation='bv', save_png='test_fpi_brst_i_bv_slice_x', display=global_display)

    def test_fpi_brst_slice_norm(self):
        mms_part_slice2d(time='2015-10-16/13:06:30', slice_norm=np.array([0, 0, 1]), probe='1', species='i', data_rate='brst', rotation='bv', save_png='test_fpi_brst_i_bv_slice_norm', display=global_display)

    def test_fpi_brst_rotations(self):
        time = '2015-10-16/13:06:30'
        # rotations = ['xy', 'xz', 'bv', 'be', 'xvel', 'perp', 'perp_xy', 'perp_xz', 'perp_yz', 'b_exb', 'perp1-perp2']
        rotations = ['bv']
        species = ['i']
        for spc in species:
            for rotation in rotations:
                mms_part_slice2d(time=time, probe='1', species=spc, data_rate='brst', rotation=rotation, save_png='test_fpi_brst_' + spc + '_' + rotation, display=global_display)

    def test_fpi_subtract_bulk(self):
        time = '2015-10-16/13:06:30'
        mms_part_slice2d(time=time, probe='1', species='i', data_rate='brst', rotation='xy',
                         interpolation='2d', save_png='test_fpi_brst_subtract_bulk',
                         subtract_bulk=True, display=global_display)

    def test_fpi_avg_angle(self):
        time = '2015-10-16/13:06:30'
        mms_part_slice2d(time=time, probe='1', species='i', data_rate='brst', rotation='xy',
                         interpolation='geometric', save_png='test_fpi_brst_avg_angle',
                         average_angle=[-45, 45], display=global_display)

    def test_fpi_sum_angle(self):
        time = '2015-10-16/13:06:30'
        mms_part_slice2d(time=time, probe='1', species='i', data_rate='brst', rotation='xy',
                         interpolation='geometric', save_png='test_fpi_brst_sum_angle',
                         sum_angle=[-45, 45], display=global_display)

    def test_fpi_energy(self):
        time = '2015-10-16/13:06:30'
        mms_part_slice2d(time=time, probe='1', species='i', data_rate='brst', rotation='xy',
                         interpolation='geometric', save_png='test_fpi_brst_energy',
                         energy=True, display=global_display)

    def test_fpi_samples(self):
        time = '2015-10-16/13:06:30'
        mms_part_slice2d(time=time, probe='1', species='i', data_rate='brst', rotation='xy',
                         interpolation='geometric', save_png='test_fpi_brst_samples',
                         samples=3, display=global_display)

    def test_fpi_window(self):
        time = '2015-10-16/13:06:30'
        mms_part_slice2d(time=time, probe='1', species='i', data_rate='brst', rotation='xy',
                         interpolation='geometric', save_png='test_fpi_brst_window',
                         window=3, display=global_display)

    def test_fpi_window_center(self):
        time = '2015-10-16/13:06:30'
        mms_part_slice2d(time=time, probe='1', species='i', data_rate='brst', rotation='xy',
                         interpolation='geometric', save_png='test_fpi_brst_window_center',
                         window=3, center_time=True, display=global_display)

    def test_fpi_custom_rotation(self):
        rot = np.zeros((3, 3))
        rot[:, 0] = [0.33796266 , -0.082956984 , 0.93749634]
        rot[:, 1] = [0.64217210 , -0.70788234 , -0.29413872]
        rot[:, 2] = [0.68803796 , 0.70144189 , -0.18596514]

        time = '2015-10-16/13:06:30'
        mms_part_slice2d(time=time, probe='1', species='i', data_rate='brst', rotation='xy',
                         interpolation='geometric', save_png='test_fpi_brst_custom_rotation',
                         custom_rotation=rot, display=global_display)

    def test_fpi_2d_interp(self):
        time = '2015-10-16/13:06:30'
        mms_part_slice2d(time=time, probe='1', species='i', data_rate='brst', rotation='xy',
                         interpolation='2d', save_png='test_fpi_brst_2d_interp', display=global_display)

    def test_fpi_2d_interp_zdirrange(self):
        time = '2015-10-16/13:06:30'
        mms_part_slice2d(time=time, probe='1', species='i', data_rate='brst', rotation='xy',
                         interpolation='2d', save_png='test_fpi_brst_zdirrange', display=global_display,
                         zdirrange=[0, 2500])

    def test_fpi_limits(self):
        time = '2015-10-16/13:06:30'
        mms_part_slice2d(time=time, probe='1', species='i', data_rate='brst', rotation='xy', erange=[0, 10000],
                         save_png='test_fpi_brst_erange', display=global_display)

    def test_fpi_electrons(self):
        time = '2015-10-16/13:06:30'
        mms_part_slice2d(time=time, probe='1', species='e', data_rate='brst', rotation='xy',
                         save_png='test_fpi_brst_electrons', display=global_display)

    def test_hpca(self):
        time = '2015-10-16/13:06:30'
        mms_part_slice2d(time=time, probe='1', species='hplus', instrument='hpca', data_rate='brst', rotation='xy',
                         save_png='test_hpca_brst', display=global_display)

    def test_hpca_trange(self):
        trange = ['2015-10-16/13:06:20', '2015-10-16/13:06:40']
        mms_part_slice2d(trange=trange, probe='1', species='hplus', instrument='hpca', data_rate='brst', rotation='xy',
                         save_png='test_hpca_brst_trange', display=global_display)

    def test_numeric_trange(self):
        from pyspedas.projects.mms.particles.mms_part_slice2d import mms_part_slice2d
        from pyspedas.tplot_tools import time_double, time_string
        time = '2017-09-10/09:32:20'
        trange = [time_double(time) - 60, time_double(time) + 60]
        # trange = ['2017-09-10/09:31:20', '2017-09-10/09:33:20']
        mms_part_slice2d(trange=trange, instrument='fpi', species='i', rotation='bv', erange=[0, 300], cmap="turbo", display=global_display, save_png='test_str_trange_turbo.png')


    def test_units_string(self):
        self.assertTrue(spd_units_string('counts') == 'Counts')
        self.assertTrue(spd_units_string('rate') == 'Rate (#/sec)')
        self.assertTrue(spd_units_string('eflux') == 'Energy Flux (eV / sec / $cm^2$ / ster / eV)')
        self.assertTrue(spd_units_string('flux') == 'Flux (# / sec / $cm^2$ / ster / eV)')
        self.assertTrue(spd_units_string('df') == 'f ($s^3$ / $cm^3$ / $km^3$)')
        self.assertTrue(spd_units_string('df_cm') == 'f ($s^3$ / $cm^6$)')
        self.assertTrue(spd_units_string('df_km') == 'f ($s^3$ / $km^6$)')
        self.assertTrue(spd_units_string('e2flux') == '$Energy^2$ Flux ($eV^2$ / sec / $cm^2$ / ster /eV)')
        self.assertTrue(spd_units_string('e3flux') == '$Energy^3$ Flux ($eV^3$ / sec / $cm^2$ / ster /eV)')


if __name__ == '__main__':
    unittest.main()



from pyspedas.projects.elfin.load import load

def state_load(trange=['2022-08-19', '2022-08-19'],
            probe='a',
            datatype='defn',
            level='l1',
            suffix='',
            get_support_data=False,
            varformat=None,
            varnames=[],
            downloadonly=False,
            notplot=False,
            no_update=False,
            time_clip=True,
            force_download=False):
    """
    This function loads data from the ELFIN State data (state)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2022-08-19', '2022-08-19']

        probe: str
            Spacecraft identifier ('a' or 'b')
            Default: 'a'

        datatype: str
            Data type.
            Valid options::

                'defn'
                'pred'

            Default: 'defn'

        level: str
            Data level; options: 'l1'
            Default: 'l1'

        suffix: str
            The tplot variable names will be given this suffix.
            Default: no suffix is added

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: only loads in data with a "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: all variables are loaded in

        varnames: list of str
            List of variable names to load
            Default: all data variables are loaded

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False

    Returns
    ----------
    list of str
        List of tplot variables created::

            ela_pos_gei
            ela_vel_gei
            ela_att_gei
            ela_att_solution_date
            ela_att_flag
            ela_att_spinper
            ela_spin_orbnorm_angle
            ela_spin_sun_angle

    Example
    -------

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> state_vars = pyspedas.projects.elfin.state(probe='a', trange=['2022-08-19', '2022-08-19'])
    >>> tplot(['ela_pos_gei', 'ela_att_gei', 'ela_att_spinper', 'ela_spin_sun_angle' ])

    """

    tvars = load(instrument='state', probe=probe, trange=trange, level=level, datatype=datatype, suffix=suffix,
                 get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly,
                 notplot=notplot, time_clip=time_clip, no_update=no_update, force_download=force_download)

    if tvars is None or notplot or downloadonly:
        return tvars

    return state_postprocessing(tvars)

def state_postprocessing(variables):
    """
    Placeholder for State post-processing
    """
    return variables




from pyspedas.utilities.datasets import find_datasets

# Refer to __init__.py for revision history before the function definitions were split out

def datasets(instrument=None, label=True):
    out = find_datasets(mission='Smallsats/Cubesats', instrument='csswe', label=label)
    return out



from .load import load, loadr
from .load_orbit import load_orbit
from functools import partial, update_wrapper
from pyspedas.utilities.pyspedas_functools import better_partial

# Define alias to the load function
orbit = load_orbit

# Define partial wrappers for other load routines, fixing the instrument parameter
# 'better_partial' works better with PyCharm autocompletion than functools.partial
# update_wrapper() is necessary for help() to show info for the wrapped function,
# rather than the partial() object.

fgm = better_partial(load, instrument="fgm")
update_wrapper(fgm, load)
eps = better_partial(load, instrument="eps")
update_wrapper(eps, load)
epead = better_partial(load, instrument="epead")
update_wrapper(epead,load)
maged = better_partial(load, instrument="maged")
update_wrapper(maged,load)
magpd = better_partial(load, instrument="magpd")
update_wrapper(magpd, load)
hepad = better_partial(load, instrument="hepad")
update_wrapper(hepad, load)
xrs = better_partial(load, instrument="xrs")
update_wrapper(xrs, load)
euvs = better_partial(load, instrument="euvs")
update_wrapper(euvs, load)
mag = better_partial(load, instrument="mag")
update_wrapper(mag, load)
mpsh = better_partial(load, instrument="mpsh")
update_wrapper(mpsh, load)
sgps = better_partial(load, instrument="sgps")
update_wrapper(sgps, load)



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG


def load(
    trange=["2007-11-01", "2007-11-02"],
    instrument="mpa",
    probe="a1",
    datatype="k0",
    prefix="",
    suffix="",
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=False,
    force_download=False,
):
    """
    Loads LANL data for the specified instrument and probe.

    Can load data from two instruments, the Magnetospheric Plasma Analyzer (mpa) and the Synchronous Orbit Particle Analyzer (spa).

    Parameters
    ----------
    trange : list of str, optional
        Time range of interest with the format ['YYYY-MM-DD','YYYY-MM-DD'] or
        to specify more or less than a day ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss'].
        Defaults to ['2004-10-31', '2004-11-01'].
    instrument : str, optional
        The instrument to load data for. Defaults to 'mpa'.
        Valid instruments: 'mpa', 'spa'.
    probe : str, optional
        The probe to load data for. Defaults to 'a1'.
        Valid probes (with gaps for some dates): 'l0', 'l1', 'l4', 'l7', 'l9', 'a1', 'a2'.
    datatype : str, optional
        The datatype. Defaults to 'k0'.
        Valid datatypes: 'k0', 'h0'.
        Data for 'h0' datatype is available only for the 'mpa' instrument and only for a few days in 1998.
    prefix : str, optional
        The tplot variable names will be given this prefix. By default, no prefix is added.
        Defaults to ''.
    suffix : str, optional
        The tplot variable names will be given this suffix. Defaults to ''.
    get_support_data : bool, optional
        Data with an attribute "VAR_TYPE" with a value of "support_data" will be loaded into tplot.
        By default, only loads in data with a "VAR_TYPE" attribute of "data". Defaults to False.
    varformat : str, optional
        The file variable formats to load into tplot. Wildcard character "*" is accepted.
        By default, all variables are loaded in. Defaults to None.
    varnames : list of str, optional
        List of variable names to load. If not specified, all data variables are loaded. Defaults to [].
    downloadonly : bool, optional
        Set this flag to download the CDF files, but not load them into tplot variables. Defaults to False.
    notplot : bool, optional
        Return the data in hash tables instead of creating tplot variables. Defaults to False.
    no_update : bool, optional
        If set, only load data from your local cache. Defaults to False.
    time_clip : bool, optional
        Time clip the variables to exactly the range specified in the trange keyword. Defaults to False.
    force_download : bool
        Download file even if local version is more recent than server version. Defaults to False.

    Returns
    -------
    list
        List of tplot variables created.
    """


    # remote directory names
    LANL_SC = {
        "l0": "90",
        "l1": "91",
        "l4": "94",
        "l7": "97",
        "l9": "89",
        "a1": "01a",
        "a2": "02a",
    }
    probe = probe.lower()

    # h0 is only available for mpa and has a different directory structure
    datatypestr = ""
    if datatype == "h0" and instrument == "mpa":
        datatypestr = "_h0"

    # path for SPDF files
    pathformat = (
        LANL_SC[probe]
        + datatypestr
        + "_"
        + instrument
        + "/%Y/"
        + probe
        + "_"
        + datatype
        + "_"
        + instrument
        + "_%Y%m%d_v??.cdf"
    )

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(
        remote_file=remote_names,
        remote_path=CONFIG["remote_data_dir"],
        local_path=CONFIG["local_data_dir"],
        no_download=no_update,
        force_download=force_download
    )
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(
        out_files,
        prefix=prefix,
        suffix=suffix,
        get_support_data=get_support_data,
        varformat=varformat,
        varnames=varnames,
        notplot=notplot,
    )

    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix="")

    return tvars
